from collections import defaultdict
from contextlib import contextmanager
from datetime import timedelta
import io
import os
import stat
from unittest.mock import MagicMock

import pytest

import confluent_kafka
from hop import models
from hop.io import StartPosition


# example GCN circular from https://gcn.gsfc.nasa.gov/gcn3_circulars.html
GCN_TITLE = "GCN GRB OBSERVATION REPORT"
GCN_NUMBER = "40"
GCN_SUBJECT = "GRB980329 VLA observations"
GCN_DATE = "98/04/03 07:10:15 GMT"
GCN_FROM = "Greg Taylor at NRAO"
GCN_BODY = """
G.B. Taylor, D.A. Frail (NRAO), S.R. Kulkarni (Caltech), and
the BeppoSAX GRB team report:

We have observed the field containing the proposed x-ray counterpart
1SAX J0702.6+3850 of GRB 980329 (IAUC 6854) with the VLA at 8.4 GHz
on UT 1998 Mar 30.2, April 1.1, and April 2.1.  Observations on April
1.1 detected a radio source VLA J0702+3850 within the 1 arcminute
error circle of 1SAX J0702.6+3850.  The coordinates of
VLA J0702+3850 are: ra = 07h02m38.02170s dec = 38d50'44.0170" (equinox
J2000) with an uncertainty of 0.05 arcsec in each coordinate.  The
size of this radio source is less than 0.25 arcsec.  The density of
sources on the sky stronger than 250 microJy at this frequency is
0.0145 arcmin**-2.

The flux density measurements of VLA J0702+3850 are as follows:

Date(UT)   8.4 GHz Flux Density
--------   ----------------------
Mar 30.2   166 +/- 50 microJy
Apr  1.1   248 +/- 16    "
Apr  2.1    65 +/- 25    "

where the uncertainty in the measurement reflects the 1 sigma rms
noise in the image.  These measurements clearly demonstrate that
the radio source is variable on timescales of less than 1 day.
This rapid variability is similar to that observed in the
radio afterglow from GRB 970508.  We propose VLA J0702+3850
is the radio afterglow from GRB 980329.

Additional radio observations are in progress.
"""

GCN_CIRCULAR = f"""\
TITLE:   {GCN_TITLE}
NUMBER:  {GCN_NUMBER}
SUBJECT: {GCN_SUBJECT}
DATE:    {GCN_DATE}
FROM:    {GCN_FROM}

{GCN_BODY}\
"""

VOEVENT_XML = """\
<?xml version='1.0' encoding='UTF-8'?>
<voe:VOEvent xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xmlns:voe="http://www.ivoa.net/xml/VOEvent/v2.0" xsi:schemaLocation="http://www.ivoa.net/xml/VOEvent/v2.0 http://www.ivoa.net/xml/VOEvent/VOEvent-v2.0.xsd" version="2.0" role="observation" ivorn="ivo://gwnet/LVC#S200302c-1-Preliminary">
  <Who>
    <Date>2020-03-02T02:00:09</Date>
    <Author>
      <contactName>LIGO Scientific Collaboration and Virgo Collaboration</contactName>
    </Author>
  </Who>
  <What>
    <Param dataType="int" name="Packet_Type" value="150">
      <Description>The Notice Type number is assigned/used within GCN, eg type=150 is an LVC_PRELIMINARY notice</Description>
    </Param>
    <Param dataType="int" name="internal" value="0">
      <Description>Indicates whether this event should be distributed to LSC/Virgo members only</Description>
    </Param>
    <Param dataType="int" name="Pkt_Ser_Num" value="1">
      <Description>A number that increments by 1 each time a new revision is issued for this event</Description>
    </Param>
    <Param dataType="string" name="GraceID" ucd="meta.id" value="S200302c">
      <Description>Identifier in GraceDB</Description>
    </Param>
    <Param dataType="string" name="AlertType" ucd="meta.version" value="Preliminary">
      <Description>VOEvent alert type</Description>
    </Param>
    <Param dataType="int" name="HardwareInj" ucd="meta.number" value="0">
      <Description>Indicates that this event is a hardware injection if 1, no if 0</Description>
    </Param>
    <Param dataType="int" name="OpenAlert" ucd="meta.number" value="1">
      <Description>Indicates that this event is an open alert if 1, no if 0</Description>
    </Param>
    <Param dataType="string" name="EventPage" ucd="meta.ref.url" value="https://gracedb.ligo.org/superevents/S200302c/view/">
      <Description>Web page for evolving status of this GW candidate</Description>
    </Param>
    <Param dataType="string" name="Instruments" ucd="meta.code" value="H1,V1">
      <Description>List of instruments used in analysis to identify this event</Description>
    </Param>
    <Param dataType="float" name="FAR" ucd="arith.rate;stat.falsealarm" unit="Hz" value="9.349090689402942e-09">
      <Description>False alarm rate for GW candidates with this strength or greater</Description>
    </Param>
    <Param dataType="string" name="Group" ucd="meta.code" value="CBC">
      <Description>Data analysis working group</Description>
    </Param>
    <Param dataType="string" name="Pipeline" ucd="meta.code" value="gstlal">
      <Description>Low-latency data analysis pipeline</Description>
    </Param>
    <Param dataType="string" name="Search" ucd="meta.code" value="AllSky">
      <Description>Specific low-latency search</Description>
    </Param>
    <Group name="GW_SKYMAP" type="GW_SKYMAP">
      <Param dataType="string" name="skymap_fits" ucd="meta.ref.url" value="https://gracedb.ligo.org/api/superevents/S200302c/files/bayestar.fits.gz,0">
        <Description>Sky Map FITS</Description>
      </Param>
    </Group>
    <Group name="Classification" type="Classification">
      <Param dataType="float" name="BNS" ucd="stat.probability" value="0.0">
        <Description>Probability that the source is a binary neutron star merger (both objects lighter than 3 solar masses)</Description>
      </Param>
      <Param dataType="float" name="NSBH" ucd="stat.probability" value="0.0">
        <Description>Probability that the source is a neutron star-black hole merger (primary heavier than 5 solar masses, secondary lighter than 3 solar masses)</Description>
      </Param>
      <Param dataType="float" name="BBH" ucd="stat.probability" value="0.8895532192171397">
        <Description>Probability that the source is a binary black hole merger (both objects heavier than 5 solar masses)</Description>
      </Param>
      <Param dataType="float" name="MassGap" ucd="stat.probability" value="0.0">
        <Description>Probability that the source has at least one object between 3 and 5 solar masses</Description>
      </Param>
      <Param dataType="float" name="Terrestrial" ucd="stat.probability" value="0.11044678078286028">
        <Description>Probability that the source is terrestrial (i.e., a background noise fluctuation or a glitch)</Description>
      </Param>
      <Description>Source classification: binary neutron star (BNS), neutron star-black hole (NSBH), binary black hole (BBH), MassGap, or terrestrial (noise)</Description>
    </Group>
    <Group name="Properties" type="Properties">
      <Param dataType="float" name="HasNS" ucd="stat.probability" value="0.0">
        <Description>Probability that at least one object in the binary has a mass that is less than 3 solar masses</Description>
      </Param>
      <Param dataType="float" name="HasRemnant" ucd="stat.probability" value="0.0">
        <Description>Probability that a nonzero mass was ejected outside the central remnant object</Description>
      </Param>
      <Description>Qualitative properties of the source, conditioned on the assumption that the signal is an astrophysical compact binary merger</Description>
    </Group>
  </What>
  <WhereWhen>
    <ObsDataLocation>
      <ObservatoryLocation id="LIGO Virgo"/>
      <ObservationLocation>
        <AstroCoordSystem id="UTC-FK5-GEO"/>
        <AstroCoords coord_system_id="UTC-FK5-GEO">
          <Time unit="s">
            <TimeInstant>
              <ISOTime>2020-03-02T01:58:11.519119</ISOTime>
            </TimeInstant>
          </Time>
        </AstroCoords>
      </ObservationLocation>
    </ObsDataLocation>
  </WhereWhen>
  <Description>Report of a candidate gravitational wave event</Description>
  <How>
    <Description>Candidate gravitational wave event identified by low-latency analysis</Description>
    <Description>V1: Virgo 3 km gravitational wave detector</Description>
    <Description>H1: LIGO Hanford 4 km gravitational wave detector</Description>
  </How>
</voe:VOEvent>\
"""

VOEVENT_JSON = """\
{"ivorn": "ivo://nasa.gsfc.gcn/SWIFT#Actual_Point_Dir_2025-07-30T17:17:19.67_1442872420-151", "role": "utility", "version": "2.0", "Who": {"AuthorIVORN": "ivo://nasa.gsfc.tan/gcn", "Author": {"shortName": "VO-GCN", "contactName": "Scott Barthelmy", "contactPhone": "+1-301-286-3106", "contactEmail": "scott.barthelmy@nasa.gov"}, "Date": "2025-07-30T17:17:14", "Description": "This VOEvent message was created with GCN VOE version: 15.08 17jun22"}, "What": {"Param": [{"name": "Packet_Type", "value": "103"}, {"name": "Pkt_Ser_Num", "value": "303"}, {"name": "TrigID", "value": "31844", "ucd": "meta.id"}, {"name": "Segment_Num", "value": "86", "ucd": "meta.id.part"}, {"name": "Slew_TJD", "value": "20886", "unit": "days", "ucd": "time"}, {"name": "Slew_SOD", "value": "62239.67", "unit": "sec", "ucd": "time"}, {"name": "Slew_RA", "value": "332.1521", "unit": "deg", "ucd": "pos.eq.ra"}, {"name": "Slew_Dec", "value": "-47.1956", "unit": "deg", "ucd": "pos.eq.dec"}, {"name": "Slew_Roll", "value": "126.2206", "unit": "deg"}, {"name": "SC_Long", "value": "74.15", "unit": "deg", "ucd": "pos.earth.lon"}, {"name": "SC_Lat", "value": "-4.83", "unit": "deg", "ucd": "pos.earth.lat"}, {"name": "Misc_flags", "value": "0x0"}, {"name": "Coords_Type", "value": "2", "unit": "dn"}, {"name": "Coords_String", "value": "pointing_direction"}], "Group": [{"name": "Misc_Flags", "Param": [{"name": "Near_Bright_Star", "value": "false"}, {"name": "Flt_Generated", "value": "true"}, {"name": "Gnd_Generated", "value": "false"}]}, {"name": "Obs_Support_Info", "Description": "The Sun and Moon values are valid at the time the VOEvent XML message was created.", "Param": [{"name": "Sun_RA", "value": "130.21", "unit": "deg", "ucd": "pos.eq.ra"}, {"name": "Sun_Dec", "value": "18.32", "unit": "deg", "ucd": "pos.eq.dec"}, {"name": "Sun_Distance", "value": "145.91", "unit": "deg", "ucd": "pos.angDistance"}, {"name": "Sun_Hr_Angle", "value": "10.51", "unit": "hr"}, {"name": "Moon_RA", "value": "195.69", "unit": "deg", "ucd": "pos.eq.ra"}, {"name": "Moon_Dec", "value": "-9.46", "unit": "deg", "ucd": "pos.eq.dec"}, {"name": "MOON_Distance", "value": "111.71", "unit": "deg", "ucd": "pos.angDistance"}, {"name": "Moon_Illum", "value": "33.28", "unit": "%", "ucd": "arith.ratio"}, {"name": "Galactic_Long", "value": "349.60", "unit": "deg", "ucd": "pos.galactic.lon"}, {"name": "Galactic_Lat", "value": "-52.46", "unit": "deg", "ucd": "pos.galactic.lat"}, {"name": "Ecliptic_Long", "value": "315.86", "unit": "deg", "ucd": "pos.ecliptic.lon"}, {"name": "Ecliptic_Lat", "value": "-33.15", "unit": "deg", "ucd": "pos.ecliptic.lat"}]}], "Description": "This is the current pointing direction for the Swift spacecraft after settling from an actual slew."}, "WhereWhen": {"ObsDataLocation": {"ObservatoryLocation": {"id": "GEOLUN"}, "ObservationLocation": {"AstroCoordSystem": {"id": "UTC-FK5-GEO"}, "AstroCoords": {"coord_system_id": "UTC-FK5-GEO", "Time": {"unit": "s", "TimeInstant": {"ISOTime": "2025-07-30T17:17:19.67"}}, "Position2D": {"unit": "deg", "Name1": "RA", "Name2": "Dec", "Value2": {"C1": "332.1521", "C2": "-47.1956"}, "Error2Radius": "0.0000"}}}}, "Description": "The RA,Dec coordinates are of the type: pointing_direction."}, "How": {"Description": "Swift Satellite", "Reference": {"uri": "http://gcn.gsfc.nasa.gov/swift.html", "type": "url"}}, "Why": {"Inference": {"probability": "1.00", "Concept": "Swift spacecraft just slewed to an observing target."}}, "Citations": {}, "Description": null, "Reference": {}}\
"""

MESSAGE_GCN_TEXT = \
    b'TITLE:            GCN/AMON NOTICE\n' \
    b'NOTICE_DATE:      Fri 12 Apr 24 05:34:35 UT\n' \
    b'NOTICE_TYPE:      ICECUBE Astrotrack Bronze \n' \
    b'STREAM:           25\n' \
    b'RUN_NUM:          139279\n' \
    b'EVENT_NUM:        10803235\n' \
    b'SRC_RA:           103.7861d {+06h 55m 09s} (J2000),\n' \
    b'                  104.1106d {+06h 56m 27s} (current),\n' \
    b'                  103.1176d {+06h 52m 28s} (1950)\n' \
    b'SRC_DEC:           +5.8716d {+05d 52\' 18"} (J2000),\n' \
    b'                   +5.8390d {+05d 50\' 20"} (current),\n' \
    b'                   +5.9364d {+05d 56\' 11"} (1950)\n' \
    b'SRC_ERROR:        69.22 [arcmin radius, stat-only, 90% containment]\n' \
    b'SRC_ERROR50:      26.96 [arcmin radius, stat-only, 50% containment]\n' \
    b'DISCOVERY_DATE:   20412 TJD;   103 DOY;   24/04/12 (yy/mm/dd)\n' \
    b'DISCOVERY_TIME:   20026 SOD {05:33:46.89} UT\n' \
    b'REVISION:         0\n' \
    b'ENERGY:           1.2126e+02 [TeV]\n' \
    b'SIGNALNESS:       3.0910e-01 [dn]\n' \
    b'FAR:              3.4096 [yr^-1]\n' \
    b'SUN_POSTN:         21.10d {+01h 24m 24s}   +8.87d {+08d 52\' 11"}\n' \
    b'SUN_DIST:          82.22 [deg]   Sun_angle= -5.5 [hr] (East of Sun)\n' \
    b'MOON_POSTN:        67.46d {+04h 29m 50s}  +26.13d {+26d 07\' 33"}\n' \
    b'MOON_DIST:         40.41 [deg]\n' \
    b'GAL_COORDS:       208.12,  3.50 [deg] galactic lon,lat of the event\n' \
    b'ECL_COORDS:       104.34,-16.88 [deg] ecliptic lon,lat of the event\n' \
    b'COMMENTS:         IceCube Bronze event.  \n' \
    b'COMMENTS:         The position error is statistical only, there is no systematic added.  '


MESSAGE_BLOB = b"This is a sample blob message. It is unstructured and does not require special parsing."

MESSAGE_JSON = b'{"foo":"bar", "baz":5}'

MESSAGE_AVRO = \
    b'\x4f\x62\x6a\x01\x04\x14\x61\x76\x72\x6f\x2e\x63' \
    b'\x6f\x64\x65\x63\x08\x6e\x75\x6c\x6c\x16\x61\x76' \
    b'\x72\x6f\x2e\x73\x63\x68\x65\x6d\x61\x8c\x0b\x7b' \
    b'\x22\x74\x79\x70\x65\x22\x3a\x20\x22\x72\x65\x63' \
    b'\x6f\x72\x64\x22\x2c\x20\x22\x6e\x61\x6d\x65\x22' \
    b'\x3a\x20\x22\x61\x75\x74\x6f\x5f\x72\x65\x63\x6f' \
    b'\x72\x64\x32\x22\x2c\x20\x22\x66\x69\x65\x6c\x64' \
    b'\x73\x22\x3a\x20\x5b\x7b\x22\x74\x79\x70\x65\x22' \
    b'\x3a\x20\x7b\x22\x74\x79\x70\x65\x22\x3a\x20\x22' \
    b'\x61\x72\x72\x61\x79\x22\x2c\x20\x22\x69\x74\x65' \
    b'\x6d\x73\x22\x3a\x20\x22\x6c\x6f\x6e\x67\x22\x7d' \
    b'\x2c\x20\x22\x6e\x61\x6d\x65\x22\x3a\x20\x22\x61' \
    b'\x72\x72\x22\x7d\x2c\x20\x7b\x22\x74\x79\x70\x65' \
    b'\x22\x3a\x20\x7b\x22\x74\x79\x70\x65\x22\x3a\x20' \
    b'\x22\x61\x72\x72\x61\x79\x22\x2c\x20\x22\x69\x74' \
    b'\x65\x6d\x73\x22\x3a\x20\x7b\x22\x74\x79\x70\x65' \
    b'\x22\x3a\x20\x22\x6d\x61\x70\x22\x2c\x20\x22\x76' \
    b'\x61\x6c\x75\x65\x73\x22\x3a\x20\x22\x6c\x6f\x6e' \
    b'\x67\x22\x7d\x7d\x2c\x20\x22\x6e\x61\x6d\x65\x22' \
    b'\x3a\x20\x22\x73\x75\x62\x5f\x6f\x62\x6a\x65\x63' \
    b'\x74\x73\x22\x7d\x2c\x20\x7b\x22\x74\x79\x70\x65' \
    b'\x22\x3a\x20\x7b\x22\x74\x79\x70\x65\x22\x3a\x20' \
    b'\x22\x72\x65\x63\x6f\x72\x64\x22\x2c\x20\x22\x6e' \
    b'\x61\x6d\x65\x22\x3a\x20\x22\x61\x75\x74\x6f\x5f' \
    b'\x72\x65\x63\x6f\x72\x64\x31\x22\x2c\x20\x22\x66' \
    b'\x69\x65\x6c\x64\x73\x22\x3a\x20\x5b\x7b\x22\x74' \
    b'\x79\x70\x65\x22\x3a\x20\x22\x73\x74\x72\x69\x6e' \
    b'\x67\x22\x2c\x20\x22\x6e\x61\x6d\x65\x22\x3a\x20' \
    b'\x22\x66\x6f\x6f\x22\x7d\x2c\x20\x7b\x22\x74\x79' \
    b'\x70\x65\x22\x3a\x20\x22\x6c\x6f\x6e\x67\x22\x2c' \
    b'\x20\x22\x6e\x61\x6d\x65\x22\x3a\x20\x22\x62\x61' \
    b'\x72\x22\x7d\x2c\x20\x7b\x22\x74\x79\x70\x65\x22' \
    b'\x3a\x20\x22\x6e\x75\x6c\x6c\x22\x2c\x20\x22\x6e' \
    b'\x61\x6d\x65\x22\x3a\x20\x22\x62\x61\x7a\x22\x7d' \
    b'\x2c\x20\x7b\x22\x74\x79\x70\x65\x22\x3a\x20\x7b' \
    b'\x22\x74\x79\x70\x65\x22\x3a\x20\x22\x72\x65\x63' \
    b'\x6f\x72\x64\x22\x2c\x20\x22\x6e\x61\x6d\x65\x22' \
    b'\x3a\x20\x22\x61\x75\x74\x6f\x5f\x72\x65\x63\x6f' \
    b'\x72\x64\x30\x22\x2c\x20\x22\x66\x69\x65\x6c\x64' \
    b'\x73\x22\x3a\x20\x5b\x7b\x22\x74\x79\x70\x65\x22' \
    b'\x3a\x20\x22\x62\x79\x74\x65\x73\x22\x2c\x20\x22' \
    b'\x6e\x61\x6d\x65\x22\x3a\x20\x22\x78\x65\x6e\x22' \
    b'\x7d\x2c\x20\x7b\x22\x74\x79\x70\x65\x22\x3a\x20' \
    b'\x7b\x22\x74\x79\x70\x65\x22\x3a\x20\x22\x61\x72' \
    b'\x72\x61\x79\x22\x2c\x20\x22\x69\x74\x65\x6d\x73' \
    b'\x22\x3a\x20\x22\x6c\x6f\x6e\x67\x22\x7d\x2c\x20' \
    b'\x22\x6e\x61\x6d\x65\x22\x3a\x20\x22\x68\x6f\x6d' \
    b'\x22\x7d\x2c\x20\x7b\x22\x74\x79\x70\x65\x22\x3a' \
    b'\x20\x22\x64\x6f\x75\x62\x6c\x65\x22\x2c\x20\x22' \
    b'\x6e\x61\x6d\x65\x22\x3a\x20\x22\x64\x72\x65\x6c' \
    b'\x22\x7d\x5d\x7d\x2c\x20\x22\x6e\x61\x6d\x65\x22' \
    b'\x3a\x20\x22\x71\x75\x75\x78\x22\x7d\x5d\x7d\x2c' \
    b'\x20\x22\x6e\x61\x6d\x65\x22\x3a\x20\x22\x74\x68' \
    b'\x69\x6e\x67\x79\x22\x7d\x2c\x20\x7b\x22\x74\x79' \
    b'\x70\x65\x22\x3a\x20\x22\x62\x79\x74\x65\x73\x22' \
    b'\x2c\x20\x22\x6e\x61\x6d\x65\x22\x3a\x20\x22\x64' \
    b'\x61\x74\x61\x22\x7d\x2c\x20\x7b\x22\x74\x79\x70' \
    b'\x65\x22\x3a\x20\x7b\x22\x74\x79\x70\x65\x22\x3a' \
    b'\x20\x22\x61\x72\x72\x61\x79\x22\x2c\x20\x22\x69' \
    b'\x74\x65\x6d\x73\x22\x3a\x20\x22\x62\x6f\x6f\x6c' \
    b'\x65\x61\x6e\x22\x7d\x2c\x20\x22\x6e\x61\x6d\x65' \
    b'\x22\x3a\x20\x22\x6c\x6f\x67\x69\x63\x22\x7d\x5d' \
    b'\x7d\x00\xfd\x2b\x62\xfc\xdf\xb1\xd2\x03\x2e\x33' \
    b'\x42\xea\xa7\x2a\xc4\x52\x02\x70\x08\x02\x04\x06' \
    b'\x08\x00\x04\x04\x02\x61\x02\x02\x62\x04\x00\x04' \
    b'\x02\x63\x06\x02\x64\x08\x00\x00\x06\x61\x62\x63' \
    b'\x2c\x06\x64\x65\x66\x06\xb2\x01\x5c\x0a\x00\x58' \
    b'\x39\xb4\xc8\x76\xbe\x05\x40\x08\x41\x00\x42\x04' \
    b'\x04\x01\x00\x00\xfd\x2b\x62\xfc\xdf\xb1\xd2\x03' \
    b'\x2e\x33\x42\xea\xa7\x2a\xc4\x52'

# Equivalent to the data encoded in MESSAGE_AVRO
AVRO_DATA_EQUIVALENT = {
    "arr": [1, 2, 3, 4],
    "sub_objects": [{"a": 1, "b": 2}, {"c": 3, "d": 4}],
    "thingy": {
        "foo": "abc",
        "bar": 22,
        "baz": None,
        "quux": {"xen": b"def", "hom": [89, 46, 5], "drel": 2.718}
    },
    "data": b"A\x00B\x04",
    "logic": [True, False],
}

GENERAL_CONFIG = """\
[config]
fetch_external = false
automatic_offload = false
"""

# This was the original configuration structure, which permitted only a single credential
AUTH_CONFIG_LEGACY = """\
[auth]
username = "username"
password = "password"
"""

# This is the new configuration structure, which contains a list of credentials
AUTH_CONFIG = """
auth = [{
         username="username",
         password="password"
         }]
"""

# This is the OIDC configuration structure, which is very similar
AUTH_CONFIG_OIDC = """
auth = [{
         username="username",
         password="password",
         token_endpoint="https://example.com/oauth2/token"
         }]
"""


class MockBroker:
    """Mock a Kafka broker.

    This stores internally messages and tracks offsets
    for different consumer groups.

    """

    def __init__(self, default_max_size=1024 * 1024):
        self._default_max_size = default_max_size
        self.reset()

    def reset(self):
        self._messages = defaultdict(list)
        self._offsets = defaultdict(dict)
        self._max_sizes = defaultdict(lambda: self._default_max_size)

    def set_default_topic_max_message_size(self, size):
        self._default_max_size = size
        new_sizes = defaultdict(lambda: self._default_max_size)
        for topic in self._max_sizes.keys():
            new_sizes[topic] = self._max_sizes[topic]
        self._max_sizes = new_sizes

    def set_topic_max_message_size(self, topic, size):
        self._max_sizes[topic] = size

    def get_topic_max_message_size(self, topic):
        return self._max_sizes[topic]

    def insert_message(self, topic, msg, headers=[], key=None):
        self._messages[topic].append((msg, headers, key))

    def write(self, topic, msg, headers=[], delivery_callback=None, key=None):
        assert delivery_callback is not None
        if len(msg) > self._max_sizes[topic]:
            if delivery_callback is None:
                raise RuntimeError("Write failed but no delivery callback set to handle the error")
            err = confluent_kafka.KafkaError(confluent_kafka.KafkaError.MSG_SIZE_TOO_LARGE,
                                             "Message too large", False, False, True)
            delivery_callback(err, None)  # TODO: synthetic message object?
            return
        self._messages[topic].append((msg, headers, key))

    def has_message(self, topic, message, headers=[], key=None):
        return (message, headers, key) in self._messages[topic]

    def read(self, topics, groupid, start_at=StartPosition.EARLIEST, **kwargs):
        if isinstance(topics, str):
            topics = [topics]

        for topic in topics:
            if topic not in self._offsets or groupid not in self._offsets[topic]:
                if start_at == StartPosition.EARLIEST:
                    self._offsets[topic][groupid] = 0
                else:
                    self._offsets[topic][groupid] = len(self._messages[topic]) - 1

            try:
                offset = self._offsets[topic][groupid]
                self._offsets[topic][groupid] += 1
                yield from self._messages[topic][offset:]
            except IndexError:
                pass


@pytest.fixture()
def mock_broker(**kwargs):
    return MockBroker(**kwargs)


class ProducerBrokerWrapper:
    def __init__(self, broker, topic):
        self.broker = broker
        self.topic = topic
        self._delay_sending = 0.0
        self._delayed = []

    def delay_sending(self, delay=1.0):
        """Simulate messages taking time to send"""
        self._delay_sending = delay

    def write(self, msg, headers=[], delivery_callback=None, topic=None, key=None):
        if topic is None:
            if self.topic is not None:
                topic = self.topic
            else:
                raise Exception("No topic specified for write")
        if self._delay_sending > 0.0:
            self._delayed.append((msg, headers, delivery_callback, topic, key))
            return
        self.broker.write(topic, msg, headers, delivery_callback=delivery_callback, key=key)

    def send_delayed(self):
        for msg, headers, delivery_callback, topic, key in self._delayed:
            self.broker.write(topic, msg, headers, delivery_callback=delivery_callback,
                              key=key)
        self._delayed.clear()

    def __len__(self):
        return len(self._delayed)

    def queued_message_count(self):
        return len(self._delayed)

    def flush(self, timeout=None):
        """Simulate time passing, possibly allowing delayed messages to send"""
        if self._delay_sending > 0.0:
            if timeout is None:
                self._delay_sending = 0.0
            else:
                self._delay_sending -= timeout.total_seconds()
                if self._delay_sending < 0.0:
                    self._delay_sending = 0.0
            if self._delay_sending == 0.0:
                self.send_delayed()
        return len(self._delayed)

    def close(self, timeout: timedelta = timedelta(seconds=0)):
        return self.flush(timeout)

    def __enter__(self):
        return self

    def __exit__(self, *exc):
        pass


@pytest.fixture()
def mock_producer():
    def _mock_producer(mock_broker, topic):
        producer = ProducerBrokerWrapper(mock_broker, topic)
        return producer

    return _mock_producer


class MockMessage:
    def __init__(self, value, headers=[]):
        self._value = value
        self._headers = headers

    def value(self):
        return self._value

    def headers(self):
        return self._headers


@pytest.fixture()
def mock_consumer():
    def _mock_consumer(mock_broker, topics, group_id, start_at=StartPosition.EARLIEST):
        class ConsumerBrokerWrapper:
            def __init__(self, broker, topics, group_id, start_at):
                if isinstance(topics, str):
                    topics = [topics]
                self.broker = broker
                self.topics = topics
                self.group_id = group_id
                self.start_at = start_at

            def subscribe(self, topics):
                if isinstance(topics, str):
                    topics = [topics]
                for topic in topics:
                    if topic not in self.topics:
                        self.topics.append(topic)

            def stream(self, *args, **kwargs):
                for message in self.broker.read(self.topics, self.group_id, self.start_at, **kwargs):
                    yield MockMessage(message[0], message[1])

            def close(self):
                pass

            def __enter__(self):
                return self

            def __exit__(self, *exc):
                pass

        consumer = ConsumerBrokerWrapper(mock_broker, topics, group_id, start_at)
        return consumer

    return _mock_consumer


@pytest.fixture()
def mock_admin_client():
    from confluent_kafka.admin import ConfigResource, ConfigEntry, ResourceType

    def _mock_admin_client(mock_broker):
        class MockAdminClient:
            def __init__(self, broker):
                self.broker = broker
                self.describe_configs_queries = 0

            def describe_configs(self, queries):
                futures = {}
                for query in queries:
                    if query.restype != ResourceType.TOPIC:
                        raise ValueError("Only TOPIC queries are emulated")
                    f = MagicMock()
                    max_message_bytes = self.broker.get_topic_max_message_size(query.name)
                    f.result = MagicMock(return_value={"max.message.bytes": ConfigEntry(
                        "max.message.bytes", max_message_bytes)})
                    futures[ConfigResource(query.restype, query.name)] = f
                    self.describe_configs_queries += 1
                return futures

            def reset_query_counter(self):
                self.describe_configs_queries = 0

        return MockAdminClient(mock_broker)

    return _mock_admin_client


@pytest.fixture(scope="session")
def mock_kafka_message():
    message = MagicMock()
    message.topic.return_value = "test-topic"
    message.partition.return_value = 0
    message.offset.return_value = 0
    message.timestamp.return_value = (0, 1234567890)
    message.key.return_value = "test-key"
    message.headers.return_value = [("a header", "a value"), ("another header", "other value")]
    return message


@pytest.fixture(scope="session")
def general_config():
    return GENERAL_CONFIG


@pytest.fixture(scope="session")
def legacy_auth_config():
    return AUTH_CONFIG_LEGACY


@pytest.fixture(scope="session")
def auth_config():
    return AUTH_CONFIG


@pytest.fixture(scope="session")
def auth_config_oidc():
    return AUTH_CONFIG_OIDC


@pytest.fixture(scope="session")
def circular_data_raw():
    return {
        "header": {
            "title": GCN_TITLE,
            "number": GCN_NUMBER,
            "subject": GCN_SUBJECT,
            "date": GCN_DATE,
            "from": GCN_FROM,
        },
        "body": GCN_BODY,
    }


@pytest.fixture(scope="session")
def circular_text():
    return GCN_CIRCULAR


@pytest.fixture(scope="session")
def circular_msg():
    return models.GCNCircular.load(GCN_CIRCULAR)


@pytest.fixture(scope="session")
def avro_data_raw():
    return AVRO_DATA_EQUIVALENT


@pytest.fixture(scope="session")
def avro_data():
    return MESSAGE_AVRO


@pytest.fixture(scope="session")
def avro_msg():
    return models.AvroBlob.load(MESSAGE_AVRO)


@pytest.fixture(scope="session")
def voevent_fileobj():
    return io.BytesIO(VOEVENT_XML.encode())


@pytest.fixture(scope="session")
def voevent_text():
    return VOEVENT_XML


@pytest.fixture(scope="session")
def gcn_text_notice_fileobj():
    return io.BytesIO(MESSAGE_GCN_TEXT)


@pytest.fixture(scope="session")
def gcn_text_notice_data():
    return MESSAGE_GCN_TEXT


@pytest.fixture(scope="session")
def blob_text():
    return MESSAGE_BLOB


@pytest.fixture(scope="session")
def blob_msg():
    return {"content": MESSAGE_BLOB}


# avro_data needs to be loaded with single_record=False here because
message_parameters_dict_data = {
    # Generalize model_name, expected_model, test_file, and model_text
    # for easy access during tests. Useful when combined with parametrization
    # across message format, since fixtures (e.g., GCN_CIRCULAR) cannot be
    # used as parametrize arguments.
    "circular": {
        "model_name": "GCNCircular",
        "expected_model": models.GCNCircular,
        "test_file": "example_gcn.gcn3",
        "model_text": GCN_CIRCULAR,
    },
    "voevent": {
        "model_name": "VOEvent",
        "expected_model": models.VOEvent,
        "test_file": "example_voevent.xml",
        "model_text": VOEVENT_XML.encode(),
    },
    "gcntextnotice": {
        "model_name": "GCNTextNotice",
        "expected_model": models.GCNTextNotice,
        "test_file": "example_gcn.txt",
        "model_text": MESSAGE_GCN_TEXT,
    },
    "json": {
        "model_name": "JSONBlob",
        "expected_model": models.JSONBlob,
        "test_file": "example_json.json",
        "model_text": MESSAGE_JSON,
    },
    "blob": {
        "model_name": "Blob",
        "expected_model": models.Blob,
        "test_file": "example_blob.txt",
        "model_text": MESSAGE_BLOB,
    },
    "avro": {
        "model_name": "AvroBlob",
        "expected_model": models.AvroBlob,
        "test_file": "example_avro.avro",
        "model_text": MESSAGE_AVRO,
    },
    # -- Outdated versions --
    "voevent-encoded-as-json": {
        "model_name": "VOEvent",
        "expected_model": models.VOEvent,
        # no test file
        "model_text": VOEVENT_JSON,
    },
}


@pytest.fixture(scope="session")
def message_parameters_dict():
    return message_parameters_dict_data


@contextmanager
def temp_environ(**vars):
    """
    A simple context manager for temporarily setting environment variables

    Kwargs:
        variables to be set and their values

    Returns:
        None
    """
    from os import environ
    original = dict(environ)
    os.environ.update(vars)
    try:
        yield  # no value needed
    finally:
        # restore original data
        os.environ.clear()
        os.environ.update(original)


@contextmanager
def temp_config(tmpdir, data):
    """
    A context manager which creates a temporary config file with specified data

    Args:
        data: the data to be written to the file

    Returns:
        The path to the config directory for hop to use this config file, as a string
    """

    config_path = f"{tmpdir}/hop/config.toml"
    os.makedirs(os.path.dirname(config_path), exist_ok=True)
    config_file = open(config_path, mode='w')
    config_file.write(data)
    config_file.close()
    try:
        yield str(tmpdir)
    finally:
        # remove file
        os.remove(config_path)


@contextmanager
def temp_auth(tmpdir, data, perms=stat.S_IRUSR | stat.S_IWUSR):
    """
    A context manager which creates a temporary auth file with specified data and permissions

    Args:
        data: the data to be written to the file
        perms: the permissions which should be set on the file.
            The default value is to use the standard, safe permissions

    Returns:
        The path to the config directory for hop to use this config file, as a string
    """

    config_path = f"{tmpdir}/hop/auth.toml"
    os.makedirs(os.path.dirname(config_path), exist_ok=True)
    config_file = open(config_path, mode='w')
    os.chmod(config_path, perms)
    config_file.write(data)
    config_file.close()
    try:
        yield str(tmpdir)
    finally:
        # remove file
        os.remove(config_path)


class PhonyConnection:
    """A mock which pretends to be a urllib3.connection.HTTPConnection well enough to fool requests
    in simple situations, allowing testing HTTP connections without actually performing any network
    requests.
    """

    def __init__(self, responses):
        self.responses = responses
        self.counter = 0
        self.requests = []

    def urlopen(self, **kwargs):
        self.requests.append(kwargs)
        if self.counter >= len(self.responses):
            raise RuntimeError("urlopen called too many times on PhonyConnection")
        response = self.responses[self.counter]
        self.counter += 1
        response.url = kwargs["url"]
        return response


def mock_pool_manager(conn: PhonyConnection):
    """Produce a mock to stand in for urllib3.poolmanager.PoolManager which returns a pre-defined
    connection object.
    """
    pm = MagicMock()
    pm.connection_from_url = MagicMock(return_value=conn)
    pm.connection_from_host = MagicMock(return_value=conn)
    return MagicMock(return_value=pm)



import builtins
from collections import OrderedDict


class HashableList(list):
    def __hash__(self):
        h = 0
        for v in self:
            h = hash(h ^ hash(v))
        return h


class HashableDict(OrderedDict):
    def __hash__(self):
        h = 0
        for k, v in self.items():
            h = hash(h ^ hash(k) ^ hash(v))
        return h


class SchemaGenerator(object):
    def __init__(self):
        # a map of visited object ids to types
        self.type_cache = {}
        self.auto_record_counter = 0

    def determine_type(self, value):
        value_id = id(value)
        if value_id in self.type_cache:
            return self.type_cache[value_id]
        type = None
        if isinstance(value, str):
            type = "string"
        elif isinstance(value, bytes):
            type = "bytes"
        elif value is None:
            type = "null"
        elif value is True:
            type = "boolean"
        elif value is False:
            type = "boolean"
        elif isinstance(value, int):
            # TODO: deal with range, int/long?
            type = "long"
        elif isinstance(value, float):
            # TODO: deal with range, float/double?
            type = "double"
        elif isinstance(value, dict):
            if not all(isinstance(k, str) for k in value.keys()):
                bad_key = next(filter(lambda k: not isinstance(k, str), value.keys()))
                raise ValueError("Dictionaries with non-string keys cannot be represented as Avro. "
                                 f"Offending key: {bad_key}")

            item_type = self.find_common_type(value.values())

            # If the resulting item type is a union, it may or may not be a good idea.
            # If the input were {"a": 1, "b": "2", "c": 3, "d": "4", . . . } with a lot of keys with
            # values of a small set of types, it may make sense, but for
            # {"name": "Bob", "age": 40, "height": 2.05}, it's mostly pointless and awkward.
            # So, we try to apply a heuristic to get this right some of the time.
            # Having a large number of value types relative to the number of entries is indicative
            # of more properly being a record (struct) type
            if isinstance(item_type, list) and len(item_type) > len(value) // 2:
                record_name = f"auto_record{self.auto_record_counter}"
                self.auto_record_counter += 1

                type = HashableDict(
                    type="record",
                    name=record_name,
                    fields=HashableList(HashableDict(name=k, type=self.determine_type(v))
                                        for k, v in value.items())
                )
            else:
                type = HashableDict(type="map", values=item_type)
        elif isinstance(value, (list, tuple)):
            item_type = self.find_common_type(value)
            type = HashableDict(type="array", items=item_type)
        else:
            raise ValueError("Unable to assign an Avro type to value of type "
                             f"{builtins.type(value)}")
        self.type_cache[value_id] = type

        return type

    def find_common_type(self, values):
        types = set()
        for value in values:
            types.add(self.determine_type(value))
        if len(types) == 1:
            return types.pop()  # only one item in set, so pop must yield it
        else:
            return HashableList(types)  # form a union type



import base64
import logging
import re
import requests
from scramp import ScramClient
import secrets
import threading


logger = logging.getLogger("hop")


class SCRAMAuth(requests.auth.AuthBase):
    def __init__(self, credential, shortcut: bool = False, check_final=True):
        """
        Args:
            credential: The set of SCRAM credentials to supply when connecting
            shortcut: Do not wait for a challenge from the server, and skip directly to sending the
                      authentication 'response', which in this case is the SCRAM
                      'client-first-message'. This saves a pointless network round-trip when
                      authentication is needed and the client can be certain that this is the
                      necessary scheme.
            check_final: Perform the final check of the server signature required by RFC 5802. The
                         value of this is unclear when communication is over HTTPS and the server
                         has already been authenticated by the TLS.
        """
        self._thread_local = threading.local()
        # these are immutable, so we do not bother making them thread-local
        self.username = credential.username
        self.password = credential.password
        self.mechanism = credential.mechanism.upper()
        self.shortcut = shortcut
        self.check_final = check_final

    def init_per_thread_state(self):
        if not hasattr(self._thread_local, "init"):
            self._thread_local.init = True
            self._thread_local.num_calls = 0
            self._thread_local.saved_body = None

    def _redo_request(self, r: requests.Response, auth_header: str, final: bool = False, **kwargs):
        # Consume content and release the original connection
        # to allow our new request to reuse the same one.
        r.content
        r.close()
        prep = r.request.copy()
        requests.cookies.extract_cookies_to_jar(prep._cookies, r.request, r.raw)
        prep.prepare_cookies(prep._cookies)
        if final and self.shortcut and self._thread_local.saved_body is not None:
            prep.prepare_body(self._thread_local.saved_body, None)

        prep.headers['Authorization'] = auth_header
        prep.register_hook("response", self.process)
        _r = r.connection.send(prep, **kwargs)
        _r.history.append(r)
        _r.request = prep
        return _r

    def _generate_client_first(self):
        """Perform the client-side preparation for the first round of the SCRAM exchange.

            Returns: The encoded data for the Authorization header
        """
        self._thread_local.nonce = secrets.token_urlsafe()
        logger.debug(f" client nonce: {self._thread_local.nonce}")
        self._thread_local.sclient = ScramClient([self.mechanism],
                                                 self.username, self.password,
                                                 c_nonce=self._thread_local.nonce)
        cfirst = self._thread_local.sclient.get_client_first()
        logger.debug(f" client first: {cfirst}")
        cfirst = cfirst.encode("utf-8")
        return f"{self.mechanism} data={base64.b64encode(cfirst).decode('utf-8')}"

    def _handle_first(self, r: requests.Response, **kwargs):
        # Need to examine which auth mechanisms the server declares it accepts to find out
        # if the one we can do is on the list
        mechanisms = requests.utils.parse_list_header(r.headers.get("www-authenticate", ""))
        matching_mech = False
        for mechanism in mechanisms:
            if mechanism.upper() == self.mechanism or \
                    mechanism.upper().startswith(self.mechanism + " "):
                matching_mech = True
                break
        if not matching_mech:
            self._thread_local.num_calls = 0
            return r
        # At this point we know our mechanism is allowed, so we begin the SCRAM exchange

        self._thread_local.num_calls = 1
        return self._redo_request(r, auth_header=self._generate_client_first(), **kwargs)

    def _handle_final(self, r: requests.Response, **kwargs):
        # To contiue the handshake, the server should have set the www-authenticate header to
        # the mechanism we are using, followed by the data we need to use.
        # Check for this, and isolate the data to parse.
        logger.debug(f"Authenticate header sent by server: {r.headers.get('www-authenticate')}")
        m = re.fullmatch(f"{self.mechanism} (.+)", r.headers.get("www-authenticate"),
                         flags=re.IGNORECASE)
        if not m:
            self._thread_local.num_calls = 0
            return r
        auth_data = requests.utils.parse_dict_header(m.group(1))
        logger.debug("auth_data:", auth_data)
        # Next, make sure that both of the fields we need were actually sent in the dictionary
        if auth_data.get("sid", None) is None:
            self._thread_local.num_calls = 0
            raise RuntimeError("Missing sid in SCRAM server first: " + m.group(1))
        if auth_data.get("data", None) is None:
            self._thread_local.num_calls = 0
            raise RuntimeError("Missing data in SCRAM server first: " + m.group(1))

        self._thread_local.sid = auth_data.get("sid")
        sfirst = auth_data.get("data")
        logger.debug(f" sid: {self._thread_local.sid}")
        sfirst = base64.b64decode(sfirst).decode("utf-8")
        logger.debug(f" server first: {sfirst}")
        self._thread_local.sclient.set_server_first(sfirst)
        cfinal = self._thread_local.sclient.get_client_final()
        logger.debug(f" client final: {cfinal}")
        cfinal = base64.b64encode(cfinal.encode("utf-8")).decode('utf-8')
        self._thread_local.num_calls = 2
        return self._redo_request(r, auth_header=f"{self.mechanism} "
                                  f"sid={self._thread_local.sid},data={cfinal}", final=True)

    def _check_server_final(self, r: requests.Response, **kwargs):
        # The standard says that we MUST authenticate the server by checking the
        # ServerSignature, and treat it as an error if they do not match.
        logger.debug(f" authentication-info: {r.headers.get('authentication-info', None)}")
        raw_auth_data = r.headers.get("authentication-info", "")
        auth_data = requests.utils.parse_dict_header(raw_auth_data)
        # Next, make sure that both of the fields we need were actually sent in the dictionary
        if auth_data.get("sid", None) is None:
            self._thread_local.num_calls = 0
            raise RuntimeError("Missing sid in SCRAM server final: " + raw_auth_data)
        if auth_data.get("data", None) is None:
            self._thread_local.num_calls = 0
            raise RuntimeError("Missing data in SCRAM server final: " + raw_auth_data)
        if auth_data.get("sid") != self._thread_local.sid:
            self._thread_local.num_calls = 0
            raise RuntimeError("sid mismatch in server final www-authenticate header")
        sfinal = auth_data.get("data")
        self._thread_local.sclient.set_server_final(base64.b64decode(sfinal).decode("utf-8"))

    def process(self, r: requests.Response, **kwargs):
        if self._thread_local.num_calls < 2 and "www-authenticate" not in r.headers:
            self._thread_local.num_calls = 0
            return r
        if self._thread_local.num_calls >= 2:
            self._check_server_final(r)
            # prevent infinite looping if something goes wrong
            r.request.deregister_hook("response", self.process)

        if r.status_code == 401:
            if self._thread_local.num_calls == 0:
                return self._handle_first(r, **kwargs)
            elif self._thread_local.num_calls == 1:
                return self._handle_final(r, **kwargs)
        return r

    def __call__(self, r):
        self.init_per_thread_state()

        r.register_hook("response", self.process)
        # This is a bit hacky and not fully general:
        # If we are using the shortcut of assuming that we must do SCRAM authentication,
        # we assume that the request will have to be repeated, so we remove the body initially and
        # squirrel it away to be re-attached to the final request at the end of the SCRAM handshake.
        # This has the advantage of not sending the potentially large body data repeatedly.
        if self.shortcut:
            self._thread_local.saved_body = r.body
            r.prepare_body(b"", None)
            r.headers["Content-Length"] = 0
            r.headers['Authorization'] = self._generate_client_first()
            self._thread_local.num_calls = 1  # skip state ahead

        return r



import logging
import sys


logger = logging.getLogger("hop")


def add_client_opts(parser):
    """Add general client options to an argument parser.

    Args:
        parser: An ArgumentParser instance to add client options to.

    """

    parser.add_argument(
        "url",
        metavar="URL",
        help="Sets the URL (kafka://host[:port]/topic) to publish messages to.",
    )
    parser.add_argument(
        "--no-auth", action="store_true", help="If set, disable authentication."
    )


def add_logging_opts(parser):
    """Add logging client options to an argument parser.

    Args:
        parser: An ArgumentParser instance to add client options to.

    """
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "-q", "--quiet", action="store_true", help="If set, only display warnings and errors."
    )
    group.add_argument(
        "-v", "--verbose", action="store_true", help="If set, display additional logging messages."
    )


def get_log_level(args):
    """Determine the log level from logging options.

    Args:
        args: The parsed argparse arguments.

    Returns:
        the logging log level.

    """
    if args.quiet:
        return logging.WARNING
    elif args.verbose:
        return logging.DEBUG
    else:
        return logging.INFO


def set_up_logger(args):
    """Set up common logging settings for CLI usage.

    Args:
        args: The parsed argparse arguments.

    """
    log_level = get_log_level(args)
    logger.setLevel(log_level)
    handler = logging.StreamHandler(sys.stderr)
    handler.setLevel(log_level)
    formatter = logging.Formatter("%(asctime)s | hop : %(levelname)s : %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)



from unittest.mock import patch, MagicMock

import pytest

from hop import auth
from hop import configure
import os
import stat
import toml

from conftest import temp_environ, temp_auth


def check_credential_file(config_path, cred):
    assert os.path.exists(config_path)
    assert os.stat(config_path).st_mode & 0o7777 == 0o600
    cf = open(config_path, "r")
    config_file_text = cf.read()
    assert cred.username in config_file_text
    assert cred.password in config_file_text
    if len(cred.hostname) > 0:
        assert cred.hostname in config_file_text


def test_auth_username():
    a = auth.Auth("foo", "bar")
    assert a.username == "foo"


def test_auth_password():
    a = auth.Auth("foo", "bar")
    assert a.password == "bar"


def test_auth_hostname():
    a = auth.Auth("foo", "bar")  # use default hostname
    assert a.hostname == ""
    a = auth.Auth("foo", "bar", "example.com")
    assert a.hostname == "example.com"


def test_auth_mechanism():
    a = auth.Auth("foo", "bar")  # use default mechanism
    assert a.mechanism == str(auth.SASLMethod.SCRAM_SHA_512)
    a = auth.Auth("foo", "bar", method=auth.SASLMethod.SCRAM_SHA_256)
    assert a.mechanism == str(auth.SASLMethod.SCRAM_SHA_256)
    a = auth.Auth("foo", "bar", token_endpoint="https://example.com/oauth2/token")
    assert a.mechanism == str(auth.SASLMethod.OAUTHBEARER)
    assert a.token_endpoint == "https://example.com/oauth2/token"


def test_auth_protocol():
    a = auth.Auth("foo", "bar")  # use default security
    assert a.ssl
    assert a.protocol == "SASL_SSL"
    a = auth.Auth("foo", "bar", ssl=False)
    assert not a.ssl
    assert a.protocol == "SASL_PLAINTEXT"


def test_auth_ca_location():
    a = auth.Auth("foo", "bar", ssl=False)
    assert not a.ssl
    assert a.ssl_ca_location is None
    a = auth.Auth("foo", "bar", ssl_ca_location="foo/bar")
    assert a.ssl
    assert a.ssl_ca_location == "foo/bar"


def test_load_auth_legacy(legacy_auth_config, tmpdir):
    with temp_auth(tmpdir, legacy_auth_config) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir):
        auth_data = auth.load_auth()
        assert len(auth_data) == 1
        assert auth_data[0].username == "username"


def test_load_auth(auth_config, tmpdir):
    with temp_auth(tmpdir, auth_config) as config_dir, temp_environ(XDG_CONFIG_HOME=config_dir):
        auth_data = auth.load_auth()
        assert len(auth_data) == 1
        assert auth_data[0].username == "username"


def test_load_auth_oidc(auth_config_oidc, tmpdir):
    with temp_auth(tmpdir, auth_config_oidc) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir):
        auth_data = auth.load_auth()
        assert len(auth_data) == 1
        assert auth_data[0].username == "username"
        assert auth_data[0].token_endpoint == "https://example.com/oauth2/token"


def test_load_auth_non_existent(auth_config, tmpdir):
    with temp_environ(XDG_CONFIG_HOME=str(tmpdir)), \
            pytest.raises(FileNotFoundError):
        auth.load_auth()


def test_load_auth_bad_perms(auth_config, tmpdir):
    for bad_perm in [stat.S_IRGRP, stat.S_IWGRP, stat.S_IXGRP,
                     stat.S_IROTH, stat.S_IWOTH, stat.S_IXOTH]:
        with temp_auth(tmpdir, auth_config, bad_perm) as config_dir, \
                temp_environ(XDG_CONFIG_HOME=config_dir), pytest.raises(RuntimeError):
            auth.load_auth()


def test_load_auth_invalid_toml(tmpdir):
    garbage = "KHFBGKJSBVJKbdfb ,s ,msb vks bs"
    with temp_auth(tmpdir, garbage) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir), pytest.raises(RuntimeError):
        auth.load_auth()


def test_load_auth_malformed_legacy(tmpdir):
    missing_username = """
                       [auth]
                       password = "password"
                       extra = "stuff"
                       """
    with temp_auth(tmpdir, missing_username) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir), pytest.raises(RuntimeError):
        auth.load_auth()

    missing_password = """
                       [auth]
                       username = "username"
                       extra = "stuff"
                   """
    with temp_auth(tmpdir, missing_password) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir), pytest.raises(RuntimeError):
        auth.load_auth()


def test_load_auth_malformed(tmpdir):
    missing_username = """
        auth = [{extra="stuff",
            password="password"}]
        """
    with temp_auth(tmpdir, missing_username) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir), pytest.raises(RuntimeError):
        auth.load_auth()

    missing_password = """
    auth = [{username="username",
        extra="stuff"}]
    """
    with temp_auth(tmpdir, missing_password) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir), pytest.raises(RuntimeError):
        auth.load_auth()


def test_load_auth_options(auth_config, tmpdir):
    # SSL should be used by default
    # The default mechanism should be SCRAM_SHA_512
    with temp_auth(tmpdir, auth_config) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir), patch("hop.auth.Auth") as auth_mock:
        auth.load_auth()
        from adc.auth import SASLMethod
        auth_mock.assert_called_with('username', 'password', host='',
                                     ssl=True, method=SASLMethod.SCRAM_SHA_512,
                                     token_endpoint=None)

    # But it should be possible to disable SSL
    use_plaintext = """auth = [{
                       username = "username",
                       password = "password",
                       protocol = "SASL_PLAINTEXT"
                       }]"""
    with temp_auth(tmpdir, use_plaintext) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir), patch("hop.auth.Auth") as auth_mock:
        auth.load_auth()
        auth_mock.assert_called_with('username', 'password', host='',
                                     ssl=False, method=SASLMethod.SCRAM_SHA_512,
                                     token_endpoint=None)

    # An SSL CA data path should be honored
    with_ca_data = """auth = [{
                   username = "username",
                   password = "password",
                   ssl_ca_location = "/foo/bar/baz"
                   }]"""
    with temp_auth(tmpdir, with_ca_data) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir), patch("hop.auth.Auth") as auth_mock:
        auth.load_auth()
        auth_mock.assert_called_with('username', 'password', host='',
                                     ssl=True, method=SASLMethod.SCRAM_SHA_512,
                                     token_endpoint=None,
                                     ssl_ca_location="/foo/bar/baz")

    # Alternate mechanisms should be honored
    plain_mechanism = """auth = [{
                      username = "username",
                      password = "password",
                      mechanism = "PLAIN"
                      }]"""
    with temp_auth(tmpdir, plain_mechanism) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir), patch("hop.auth.Auth") as auth_mock:
        auth.load_auth()
        auth_mock.assert_called_with('username', 'password', host='',
                                     ssl=True, method=SASLMethod.PLAIN,
                                     token_endpoint=None)

    # Associated hostnames should be included
    with_host = """auth = [{
                username = "username",
                password = "password",
                hostname = "example.com"
                }]"""
    with temp_auth(tmpdir, with_host) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir):
        creds = auth.load_auth()
        assert len(creds) == 1
        assert creds[0].hostname == "example.com"


def test_load_auth_muliple_creds(tmpdir):
    two_creds = """auth = [
                    {
                        username = "user1",
                        password = "pass1",
                        hostname = "host1"
                    },
                    {
                        username = "user2",
                        password = "pass2",
                        hostname = "host2"
                    },
                ]"""
    with temp_auth(tmpdir, two_creds) as config_dir, temp_environ(XDG_CONFIG_HOME=config_dir):
        creds = auth.load_auth()
        assert len(creds) == 2
        assert creds[0].username == "user1"
        assert creds[0].password == "pass1"
        assert creds[0].hostname == "host1"
        assert creds[1].username == "user2"
        assert creds[1].password == "pass2"
        assert creds[1].hostname == "host2"


def check_no_auth_data(file_path):
    """Check that a config file contains no auth data, but is otherwise valid TOML

    Returns:
        The configuration which was read from the file
    """
    with open(file_path, "r") as f:
        config_data = toml.loads(f.read())
        assert "auth" not in config_data
        return config_data


def test_prune_outdated_empty(tmpdir):
    with temp_environ(XDG_CONFIG_HOME=str(tmpdir)):
        # when the config file does not exist, this function should successfully do nothing
        auth.prune_outdated_auth()

        config_path = configure.get_config_path()
        os.makedirs(os.path.dirname(config_path), exist_ok=True)
        with open(config_path, "w"):
            pass
        # should also work when the file exists but is empty
        auth.prune_outdated_auth()
        check_no_auth_data(config_path)


def test_prune_outdated_malformed(tmpdir):
    with temp_environ(XDG_CONFIG_HOME=str(tmpdir)):
        config_path = configure.get_config_path()
        os.makedirs(os.path.dirname(config_path), exist_ok=True)
        with open(config_path, "w") as f:
            f.write("not valid TOML IGIUF T J2(YHFOh q3pi8hoU *AHou7w3ht")
        # a RuntimeError should be raised when the file is unparseable garbage
        with pytest.raises(RuntimeError) as err:
            auth.prune_outdated_auth()
        assert f"configuration file {config_path} is malformed" in err.value.args[0]


def test_prune_outdated_auth(tmpdir):
    config_data = """
        foo = "bar"
        auth = [{
        username = "username",
        password = "password",
        hostname = "example.com"
        }]
        baz = "quux"
        """
    with temp_environ(XDG_CONFIG_HOME=str(tmpdir)):
        config_path = configure.get_config_path()
        os.makedirs(os.path.dirname(config_path), exist_ok=True)
        with open(config_path, "w") as f:
            f.write(config_data)
        auth.prune_outdated_auth()
        # the auth data should be gone
        new_config_data = check_no_auth_data(config_path)
        # all other data should remain untouched
        assert "foo" in new_config_data
        assert "baz" in new_config_data

        # the same should work for files in non-default locations
        alt_config_path = f"{os.path.dirname(config_path)}/other.toml"
        with open(alt_config_path, "w") as f:
            f.write(config_data)
        auth.prune_outdated_auth(alt_config_path)
        # the auth data should be gone
        new_config_data = check_no_auth_data(alt_config_path)
        # all other data should remain untouched
        assert "foo" in new_config_data
        assert "baz" in new_config_data


def test_select_auth_inexact_host():
    choices = [auth.Auth("user1", "pass1"),  # no host
               auth.Auth("user2", "pass2", host="some_host"),
               auth.Auth("user3", "pass3", host="other_host:9092")]
    # A credential with no host specified should match any host not exactly matched by any other
    # credential's host, when no username is specified
    selected = auth.select_matching_auth(choices, "example.com", None)
    assert selected == choices[0]


def test_select_auth_no_match(auth_config, tmpdir):
    no_match = "No matching credential found"

    # no credentials at all
    with pytest.raises(RuntimeError) as err:
        selected = auth.select_matching_auth([], "example.com", "nosuchuser")
    assert "No matching credential found for hostname 'example.com'" in err.value.args[0]

    # no match for requested hostname
    with_host = """auth = [{
        username = "username",
        password = "password",
        hostname = "example.com"
        }]"""
    with temp_auth(tmpdir, with_host) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir), pytest.raises(RuntimeError) as err:
        creds = auth.load_auth()
        selected = auth.select_matching_auth(creds, "example.net")
    assert f"{no_match} for hostname 'example.net'" in err.value.args[0]

    # no match for requested username
    with temp_auth(tmpdir, auth_config) as config_dir, \
            temp_environ(XDG_CONFIG_HOME=config_dir), pytest.raises(RuntimeError) as err:
        creds = auth.load_auth()
        selected = auth.select_matching_auth(creds, "example.com", "nosuchuser")
    assert f"{no_match} for hostname 'example.com' with username 'nosuchuser'" in err.value.args[0]


def test_select_auth_ambiguity():
    too_many = "Ambiguous credentials found"

    two_vague_creds = [
        auth.Auth("user1", "pass1"),
        auth.Auth("user2", "pass2"),
    ]

    # given two credetials with no associated hostnames, any lookup which does not specify the
    # username will be ambiguous
    with pytest.raises(RuntimeError) as err:
        selected = auth.select_matching_auth(two_vague_creds, "example.com")
    assert f"{too_many} for hostname 'example.com' with no username specified" in err.value.args[0]
    assert "user1 which has no associated hostname" in err.value.args[0]
    assert "user2 which has no associated hostname" in err.value.args[0]

    # specifying a valid username should resolve the ambiguity
    for username in ["user1", "user2"]:
        selected = auth.select_matching_auth(two_vague_creds, "example.com", username)
        assert selected.username == username

    two_vague_creds = [
        auth.Auth("user1", "pass1"),
        auth.Auth("user3", "pass3", host="example.com"),
    ]
    # given a credential with no hostname and one with, a request which matches the one with a
    # hostname should _not_ be ambiguous
    selected = auth.select_matching_auth(two_vague_creds, "example.com")
    assert selected.username == "user3"

    two_specific_creds = [
        auth.Auth("user3", "pass3", host="example.com"),
        auth.Auth("user4", "pass4", host="example.com"),
    ]

    # given two credentials which both exactly match the requested hostname there should again be
    # an ambiguity
    with pytest.raises(RuntimeError) as err:
        selected = auth.select_matching_auth(two_specific_creds, "example.com")
    assert f"{too_many} for hostname 'example.com'" in err.value.args[0]
    assert "user3 for example.com" in err.value.args[0]
    assert "user4 for example.com" in err.value.args[0]

    # specifying a valid username should again resolve the ambiguity
    for username in ["user3", "user4"]:
        selected = auth.select_matching_auth(two_specific_creds, "example.com", username)
        assert selected.username == username

    # No single source of credentials should issue two credentials with the same username, however,
    # two separate issuers could issue ones with the same name, and if neither specifies a hostname
    # they will be ambiguous.
    # (Duplicate usernames with matching hostnames should be impossible, as there should not be two
    # issuers for the same hostname)
    duplicate_users = [
        auth.Auth("user", "pass5"),
        auth.Auth("user", "pass6"),
    ]
    with pytest.raises(RuntimeError) as err:
        selected = auth.select_matching_auth(duplicate_users, "example.com")
    assert f"{too_many} for hostname 'example.com'" in err.value.args[0]
    assert "user which has no associated hostname" in err.value.args[0]
    assert "user which has no associated hostname" in err.value.args[0]

    # specifying a username woun't help, since they are duplicates
    with pytest.raises(RuntimeError) as err:
        selected = auth.select_matching_auth(duplicate_users, "example.com", "user")
    assert f"{too_many} for hostname 'example.com'" in err.value.args[0]
    assert "user which has no associated hostname" in err.value.args[0]
    assert "user which has no associated hostname" in err.value.args[0]


def test_select_auth_default_port():
    choices = [auth.Auth("user", "pass", host="host"),
               auth.Auth("user", "pass", host="other_host:9092")]
    # A target with the default port explicitly specified should match a credential with the default
    # port left implicit.
    selected = auth.select_matching_auth(choices, "host:9092")
    assert selected == choices[0]

    choices = [auth.Auth("user", "pass", host="host:9092"),
               auth.Auth("user", "pass", host="other_host")]
    # A target with the port implicitly defaulted should match a credential with the default
    # port left explicitly specified.
    selected = auth.select_matching_auth(choices, "host")
    assert selected == choices[0]


def test_auth_location_fallback(tmpdir):
    valid_auth = "auth = [{username=\"user\",password=\"pass\"}]"
    other_auth = "auth = [{username=\"other-user\",password=\"other-pass\"}]"

    config_dir = f"{tmpdir}/hop"
    os.makedirs(config_dir, exist_ok=True)

    def write_file(name: str, data: str):
        file_path = f"{config_dir}/{name}"
        with open(file_path, 'w') as file:
            os.chmod(file_path, stat.S_IRUSR | stat.S_IWUSR)
            file.write(data)

    with temp_environ(XDG_CONFIG_HOME=str(tmpdir)):
        # If both auth.toml and config.toml contain auth data, only auth.toml should be read
        write_file("auth.toml", valid_auth)
        write_file("config.toml", other_auth)
        creds = auth.load_auth()
        assert len(creds) == 1
        assert creds[0].username == "user"

        # If auth.toml does not exist and config.toml contains valid auth data, it should be read
        os.remove(f"{config_dir}/auth.toml")
        creds = auth.load_auth()
        assert len(creds) == 1
        assert creds[0].username == "other-user"

        # If auth.toml does not exist and config.toml exists but contains no valid auth data, the
        # resulting error should be about auth.toml
        write_file("config.toml", "")
        with pytest.raises(FileNotFoundError) as err:
            creds = auth.load_auth()
        assert "auth.toml" in err.value.filename

        # If neither auth.toml nor config.toml exixts, the error should be about auth.toml
        os.remove(f"{config_dir}/config.toml")
        with pytest.raises(FileNotFoundError) as err:
            creds = auth.load_auth()
        assert "auth.toml" in err.value.filename


def test_read_new_credential_csv(tmpdir):
    csv_file = tmpdir + "/cred.csv"

    # read from a csv file with no hostname
    with open(csv_file, "w") as f:
        f.write("username,password\n")
        f.write("user,pass")
    new_cred = auth.read_new_credential(csv_file)
    assert new_cred.username == "user"
    assert new_cred.password == "pass"
    assert new_cred.hostname == ""

    # read from a csv file with a hostname
    with open(csv_file, "w") as f:
        f.write("username,password,hostname\n")
        f.write("user2,pass2,example.com")
    new_cred = auth.read_new_credential(csv_file)
    assert new_cred.username == "user2"
    assert new_cred.password == "pass2"
    assert new_cred.hostname == "example.com"

    # read from a csv file with a mechanism
    with open(csv_file, "w") as f:
        f.write("username,password,mechanism\n")
        f.write("user3,pass3,SCRAM-SHA-256")
    new_cred = auth.read_new_credential(csv_file)
    assert new_cred.username == "user3"
    assert new_cred.password == "pass3"
    assert new_cred.mechanism == "SCRAM_SHA_256"

    # read from a csv file with a mechanism
    with open(csv_file, "w") as f:
        f.write("username,password,protocol\n")
        f.write("user4,pass4,SASL_PLAINTEXT")
    new_cred = auth.read_new_credential(csv_file)
    assert new_cred.username == "user4"
    assert new_cred.password == "pass4"
    assert not new_cred.ssl
    assert new_cred.protocol == "SASL_PLAINTEXT"

    # read from a csv file with an SSL CA data location
    with open(csv_file, "w") as f:
        f.write("username,password,ssl_ca_location\n")
        f.write("user5,pass5,foo/bar")
    new_cred = auth.read_new_credential(csv_file)
    assert new_cred.username == "user5"
    assert new_cred.password == "pass5"
    assert new_cred.ssl_ca_location == "foo/bar"


def test_read_new_credential_csv_malformed(tmpdir):
    csv_file = tmpdir + "/cred.csv"

    # nonexistent file => FileNotFoundError
    with pytest.raises(FileNotFoundError):
        auth.read_new_credential(csv_file)

    # no username => KeyError
    with open(csv_file, "w") as f:
        f.write("password,hostname\n")
        f.write("pass,example.com")
    with pytest.raises(KeyError):
        auth.read_new_credential(csv_file)

    # no password => KeyError
    with open(csv_file, "w") as f:
        f.write("username,hostname\n")
        f.write("user,example.com")
    with pytest.raises(KeyError):
        auth.read_new_credential(csv_file)


def test_read_new_credential_interactive(tmpdir):
    username = "foo"
    password = "bar"
    for hostname in ["", "example.com"]:
        with patch("getpass.getpass", MagicMock(return_value=password)), \
                patch("hop.auth.input", MagicMock(side_effect=[username, hostname, ""])):
            new_cred = auth.read_new_credential()
            assert new_cred.username == username
            assert new_cred.password == password
            assert new_cred.hostname == hostname


def test_read_new_credential_interactive_invalid(tmpdir):
    username = "foo"
    password = "bar"
    hostname = "example.com"

    # missing username
    with patch("getpass.getpass", MagicMock(return_value=password)), \
            patch("hop.auth.input", MagicMock(side_effect=["", hostname, ""])), \
            pytest.raises(RuntimeError) as err:
        auth.read_new_credential()
    assert err.value.args[0] == "Username may not be empty"

    # missing password
    with patch("getpass.getpass", MagicMock(return_value="")), \
            patch("hop.auth.input", MagicMock(side_effect=[username, hostname, ""])), \
            pytest.raises(RuntimeError) as err:
        auth.read_new_credential()
    assert err.value.args[0] == "Password may not be empty"


def test_hostname_validation():
    username = "foo"
    password = "bar"
    # good cases
    for hostname in ["", "example.com", "example.com:9092",
                     "kafka://example.com", "kafka://example.com:9092",
                     "127.0.0.1", "127.0.0.1:9092",
                     "[1:2::3]", "[1:2::3]:9092"]:
        processed_hostname = auth._validate_hostname(hostname)
        if "kafka://" in hostname:
            assert processed_hostname == hostname[8:]
        else:
            assert processed_hostname == hostname
    # bad cases
    for hostname in ["http://example.com", "file:///usr/bin", "foo/bar",
                     "kafka://example.com/topic",
                     "kafka://example.com:9092/topic"]:
        with pytest.raises(RuntimeError) as err:
            processed_hostname = auth._validate_hostname(hostname)
        assert "Unable to parse hostname" in err.value.args[0]


def credential_write_read_round_trip(orig_cred, file_path):
    auth.write_auth_data(file_path, [orig_cred])
    read_creds = auth.load_auth(file_path)
    assert len(read_creds) == 1
    assert read_creds[0] == orig_cred


def test_write_config_data(tmpdir):
    config_file = tmpdir + "/config"
    username = "scimma"
    password = "scimmapass"
    auth.write_auth_data(config_file, [auth.Auth(username, password)])
    check_credential_file(config_file, auth.Auth(username, password))

    credential_write_read_round_trip(auth.Auth(username, password), config_file)
    credential_write_read_round_trip(auth.Auth(username, password, host="example.com"), config_file)
    credential_write_read_round_trip(auth.Auth(username, password, ssl=False), config_file)
    credential_write_read_round_trip(auth.Auth(username, password,
                                               method=auth.SASLMethod.SCRAM_SHA_256),
                                     config_file)
    credential_write_read_round_trip(auth.Auth(username, password, ssl_ca_location="ca.cert"),
                                     config_file)


def test_list_credentials(tmpdir, capsys):
    with patch("hop.auth.load_auth", MagicMock(return_value=[])):
        auth.list_credentials()
        captured = capsys.readouterr()
        assert len(captured.out) == 0

    # lie about output going to a TTY to check more user-friendly message used then
    with patch("hop.auth.load_auth", MagicMock(return_value=[])), \
            patch("os.isatty", MagicMock(return_value=True)):
        auth.list_credentials()
        captured = capsys.readouterr()
        assert "No credentials" in captured.out

    short_cred = auth.Auth("user1", "pass1")
    long_cred = auth.Auth("user2", "pass2", "host2")

    with patch("hop.auth.load_auth", MagicMock(return_value=[short_cred])):
        auth.list_credentials()
        captured = capsys.readouterr()
        assert short_cred.username in captured.out
        assert short_cred.password not in captured.out
        assert "for" not in captured.out

    with patch("hop.auth.load_auth", MagicMock(return_value=[short_cred, long_cred])):
        auth.list_credentials()
        captured = capsys.readouterr()
        assert short_cred.username in captured.out
        assert long_cred.username in captured.out
        assert long_cred.password not in captured.out
        assert long_cred.hostname in captured.out


def test_add_credential_to_empty(tmpdir):
    new_cred = auth.Auth("user", "pass")

    with temp_environ(HOME=str(tmpdir)), \
            patch("hop.auth.read_new_credential", MagicMock(return_value=new_cred)):
        args = MagicMock()
        args.cred_file = None
        args.force = False
        auth.add_credential(args)
        check_credential_file(configure.get_config_path("auth"), new_cred)


def test_add_credential_to_nonempty(auth_config, tmpdir):
    old_cred = auth.Auth("username", "password")
    new_cred = auth.Auth("other_user", "other_pass")

    with temp_auth(tmpdir, auth_config) as config_dir, temp_environ(HOME=config_dir), \
            patch("hop.auth.read_new_credential", MagicMock(return_value=new_cred)):
        args = MagicMock()
        args.cred_file = None
        args.force = False
        auth.add_credential(args)
        check_credential_file(configure.get_config_path("auth"), old_cred)
        check_credential_file(configure.get_config_path("auth"), new_cred)


def test_add_credential_to_nonempty_no_hostname_no_conflict(tmpdir):
    # unfortunately, we must permit duplicate usernames if one has a hostname and the other does not
    old_cred = auth.Auth("username", "password", "example.com")
    new_cred = auth.Auth("username", "other_pass")

    with temp_environ(HOME=str(tmpdir)), \
            patch("hop.auth.load_auth", MagicMock(return_value=[old_cred])), \
            patch("hop.auth.read_new_credential", MagicMock(return_value=new_cred)):
        args = MagicMock()
        args.cred_file = None
        args.force = False
        auth.add_credential(args)
        check_credential_file(configure.get_config_path("auth"), old_cred)
        check_credential_file(configure.get_config_path("auth"), new_cred)


def test_add_credential_to_nonempty_hostname_no_conflict(tmpdir):
    # unfortunately, we must permit duplicate usernames if one has a hostname and the other does not
    old_cred = auth.Auth("username", "password")
    new_cred = auth.Auth("username", "other_pass", "example.com")

    with temp_environ(HOME=str(tmpdir)), \
            patch("hop.auth.load_auth", MagicMock(return_value=[old_cred])), \
            patch("hop.auth.read_new_credential", MagicMock(return_value=new_cred)):
        args = MagicMock()
        args.cred_file = None
        args.force = False
        auth.add_credential(args)
        check_credential_file(configure.get_config_path("auth"), old_cred)
        check_credential_file(configure.get_config_path("auth"), new_cred)


def test_add_credential_conflict_no_host(tmpdir, caplog):
    old_cred = auth.Auth("username", "password")
    new_cred = auth.Auth("username", "other_pass")

    with temp_environ(HOME=str(tmpdir)), \
            patch("hop.auth.read_new_credential", MagicMock(return_value=new_cred)):
        auth.write_auth_data(configure.get_config_path("auth"), [old_cred])
        args = MagicMock()
        args.cred_file = None
        args.force = False
        auth.add_credential(args)
        # without the force option, the old credential should not be overwritten
        check_credential_file(configure.get_config_path("auth"), old_cred)
        assert "Credential already exists; overwrite with --force" in caplog.text

        args.force = True
        auth.add_credential(args)
        # with the force option, the old credential should be overwritten
        check_credential_file(configure.get_config_path("auth"), new_cred)


def test_add_credential_conflict_with_host(tmpdir, caplog):
    old_cred = auth.Auth("username", "password", "example.com")
    new_cred = auth.Auth("username", "other_pass", "example.com")
    unrelated_cred = auth.Auth("username", "unrelated_pass", "example.org")

    with temp_environ(HOME=str(tmpdir)), \
            patch("hop.auth.read_new_credential", MagicMock(return_value=new_cred)):
        auth.write_auth_data(configure.get_config_path("auth"), [old_cred, unrelated_cred])
        args = MagicMock()
        args.cred_file = None
        args.force = False
        auth.add_credential(args)
        # without the force option, the old credential should not be overwritten
        check_credential_file(configure.get_config_path("auth"), old_cred)
        assert "Credential already exists; overwrite with --force" in caplog.text

        args.force = True
        auth.add_credential(args)
        # with the force option, the old credential should be overwritten
        check_credential_file(configure.get_config_path("auth"), new_cred)
        # also check that unrelated credentials haven't been removed
        creds = auth.load_auth(configure.get_config_path("auth"))
        assert unrelated_cred in creds


delete_input_creds = [
    auth.Auth("user1", "pass1"),
    auth.Auth("user2", "pass2"),
    auth.Auth("user3", "pass3", "example.com"),
    auth.Auth("user3", "pass3-alt", "example.net"),
]


def check_credential_deletion(config_file, removed_index):
    """Check that a particular credential is not in a config file after removal,
       and that all other are still present.

    Args:
        config_file: Path to the file to check.
        removed_index: Index of the credential in delete_input_creds which should not be present.
    """
    observed = auth.load_auth(config_file)
    for idx, cred in enumerate(delete_input_creds):
        if idx == removed_index:
            assert cred not in observed
        else:
            assert cred in observed


def test_delete_credential_no_match(tmpdir):
    invalid_username = "not-a-user"
    with temp_environ(HOME=str(tmpdir)), \
            patch("hop.auth.load_auth", MagicMock(return_value=delete_input_creds.copy())), \
            pytest.raises(RuntimeError) as err:
        auth.delete_credential(invalid_username)
    assert "No matching credential found" in err.value.args[0]
    assert invalid_username in err.value.args[0]

    invalid_host = "example.com"
    with temp_environ(HOME=str(tmpdir)), \
            patch("hop.auth.load_auth", MagicMock(return_value=delete_input_creds.copy())), \
            pytest.raises(RuntimeError) as err:
        auth.delete_credential(f"{invalid_username}@{invalid_host}")
    assert "No matching credential found" in err.value.args[0]
    assert invalid_username in err.value.args[0]
    assert f" with hostname '{invalid_host}'" in err.value.args[0]


def test_delete_credential_no_match_no_data(tmpdir):
    invalid_username = "not-a-user"
    with temp_environ(HOME=str(tmpdir)), \
            pytest.raises(RuntimeError) as err:
        auth.delete_credential(invalid_username)
    assert "No matching credential found" in err.value.args[0]
    assert invalid_username in err.value.args[0]


def test_delete_credential_no_hostname(tmpdir):
    with temp_environ(HOME=str(tmpdir)):
        with patch("hop.auth.load_auth", MagicMock(return_value=delete_input_creds.copy())):
            auth.delete_credential("user1")
        check_credential_deletion(configure.get_config_path("auth"), 0)


def test_delete_credential_with_hostname(tmpdir):
    with temp_environ(HOME=str(tmpdir)):
        with patch("hop.auth.load_auth", MagicMock(return_value=delete_input_creds.copy())):
            auth.delete_credential("user3@example.com")
        check_credential_deletion(configure.get_config_path("auth"), 2)


def test_delete_credential_ambiguous(tmpdir):
    with temp_environ(HOME=str(tmpdir)), \
            patch("hop.auth.load_auth", MagicMock(return_value=delete_input_creds.copy())), \
            pytest.raises(RuntimeError) as err:
        auth.delete_credential("user3")
    assert "Ambiguous credentials found" in err.value.args[0]
    assert delete_input_creds[2].username in err.value.args[0]
    assert delete_input_creds[2].hostname in err.value.args[0]
    assert delete_input_creds[3].hostname in err.value.args[0]


def test_delete_credential_ambiguous_with_host(tmpdir):
    creds = [delete_input_creds[2], delete_input_creds[2]]
    with temp_environ(HOME=str(tmpdir)), \
            patch("hop.auth.load_auth", MagicMock(return_value=creds)), \
            pytest.raises(RuntimeError) as err:
        auth.delete_credential("user3@example.com")
    assert "Ambiguous credentials found" in err.value.args[0]
    assert creds[0].username in err.value.args[0]
    assert creds[0].hostname in err.value.args[0]


def test_delete_credential_ambiguous_creds_without_hosts(tmpdir):
    creds = [auth.Auth("user1", "pass1"), auth.Auth("user1", "pass2")]
    with temp_environ(HOME=str(tmpdir)), \
            patch("hop.auth.load_auth", MagicMock(return_value=creds)), \
            pytest.raises(RuntimeError) as err:
        auth.delete_credential("user1")
    assert "Ambiguous credentials found" in err.value.args[0]
    assert creds[0].username in err.value.args[0]



import heapq
from io import BytesIO
import logging
import os
import struct
import time
import threading
from typing import Union
import zlib

from . import io
from .auth import select_matching_auth

import confluent_kafka
from adc import kafka
from adc.errors import KafkaException


logger = logging.getLogger("hop")


class _RAPriorityQueue:
    """A priority queue which also allows random access to queued items.

    Items' keys are also their priorities, and keys which compare lower have higher priority.

    All keys in a queue must be mutually comparable; this is most easily accomplished by using a
    single key type for a given queue.

    """

    def __init__(self):
        """Create an empty queue"""
        self.priorities = []
        self.data = {}

    def __len__(self):
        """Return: the number of items in the queue."""
        assert len(self.data) == len(self.priorities)
        return len(self.data)

    def __contains__(self, key):
        """Test whether an item with the given key/priority is present in the queue."""
        return key in self.data

    def insert(self, key, value):
        """Add an item to the queue, or replace the existing item stored under the given key."""
        self.data[key] = value
        heapq.heappush(self.priorities, key)

    def __setitem__(self, key, value):
        """Add an item to the queue, or replace the existing item stored under the given key."""
        self.insert(key, value)

    def __getitem__(self, key):
        """Fetch the item stored under the given key.

        Raises: KeyError if the key does not exist.
        """
        return self.data[key]

    def pop_highest_priority(self):
        """Fetch the item with the highest priority (lowest key) currently in the queue,
            and removes it from the queue.

        Returns: The (key, value) tuple for the highest priority item in the queue, or None if the
                 queue is empty.
        """
        if len(self.data) == 0:
            return None
        key = heapq.heappop(self.priorities)
        value = self.data[key]
        del self.data[key]
        return (key, value)

    def remove(self, key):
        """Remove an item from the queue, regardless of its location.

        Raises: KeyError if the key does not exist.
        """
        del self.data[key]
        idx = self.priorities.index(key)
        del self.priorities[idx]
        heapq.heapify(self.priorities)

    def __delitem__(self, key):
        self.remove(key)


def _ensure_bytes_like(thing):
    """Force an object which may be string-like to be bytes-like

    Args:
        thing: something which might be a string or might already be encoded as bytes.

    Return:
        Either the original input object or the encoding of that object as bytes.
    """
    try:  # check whether thing is bytes-like
        memoryview(thing)
        return thing  # keep as-is
    except TypeError:
        return thing.encode("utf-8")


class PublicationJournal:
    """
    An object which tracks the state of messages which are being sent, persists that state to disk,
    and enables it to be restored if the program stops unexpectedly.
    """

    int_format = "!Q"
    int_size = struct.calcsize(int_format)
    crc_format = "!I"
    crc_size = struct.calcsize(crc_format)
    # The original record type supported only the message body and headers
    msg_record_type_v1 = 0
    # The second version of the record type adds support for the destination topic
    # and a message key
    msg_record_type_v2 = 2
    sent_record_type = 1

    @staticmethod
    def encode_int(value: int):
        return struct.pack(PublicationJournal.int_format, value)

    @staticmethod
    def decode_int(buffer: bytes):
        return struct.unpack(PublicationJournal.int_format, buffer)[0]

    @staticmethod
    def encode_crc(crc: int):
        return struct.pack(PublicationJournal.crc_format, crc)

    @staticmethod
    def decode_crc(buffer: bytes):
        return struct.unpack(PublicationJournal.crc_format, buffer)[0]

    def __init__(self, journal_path="publisher.journal"):
        """Prepare a journal, including loading any data previously persisted to disk.

        Args:
            journal_path: The filesystem path from/to which the journal data should be read/written.

        Raises:
            PermissionError: If existing journal file does not have suitable permissions.
            RuntimeError: If existing journal data cannot be read.
        """
        self.journal_path = journal_path
        # each queue will store (message, headers) tuples, keyed by sequence numbers
        self.messages_to_send = _RAPriorityQueue()
        self.maybe_sent_messages = _RAPriorityQueue()
        self.inflight_topics = {}
        self.message_counter = 0
        self._read_previous_journal()
        self.journal = open(self.journal_path, "ab")

    def _write_record(self, record_type: int, record_body: bytes):
        try:
            body_crc = zlib.crc32(record_body, 0) & 0xFFFFFFFF
            header_crc = 0

            def write_to_header(data: bytes):
                # update running checksum and write data
                nonlocal header_crc
                header_crc = zlib.crc32(data, header_crc) & 0xFFFFFFFF
                self.journal.write(data)
            # The header is a fixed length, so it can be read independent of its contents.
            # The body length is contained in the header, and protected by the header CRC
            # so that item lengths within the body can be sanity checked before reading the
            # body is complete. The body CRC also protects the data in the body, but cannot
            # be re-checked until the entire body has been read. That can pose a problem if
            # the length for one of the variable length components in the body is corrupted,
            # and would lead to read of a nonsensical length (exhausting memory or similar).
            write_to_header(PublicationJournal.encode_int(record_type))  # record type
            write_to_header(PublicationJournal.encode_int(len(record_body)))  # body length
            write_to_header(PublicationJournal.encode_crc(body_crc))  # body CRC
            self.journal.write(PublicationJournal.encode_crc(header_crc))  # header CRC
            self.journal.write(record_body)  # body
            self.journal.flush()
        except BaseException as e:
            raise RuntimeError(f"Failed to append record to journal: {e}")

    # returns the sequence number assigned to the message
    def queue_message(self, message: bytes, topic: str, headers=None,
                      key: Union[bytes, str] = None):
        """Record to the journal a message which is to be sent.

        Args:
            message: A message to send, encoded as a bytes-like object.
            topic: The destination topic for the message.
            headers: Headers to be sent with the message, as a list of 2-tuples of bytes-like
                     objects.
            key: The key, if any to be used for the message.

        Returns:
            The sequence number assigned to the message.
            Sequence numbers are unique among all messages which are 'live' at the same time,
            but will otherwise be recycled.

        Raises:
            RuntimeError: If appending the new message to the on-disk journal fails.
            TypeError: If the message is not a suitable type (bytes)
        """
        message = _ensure_bytes_like(message)
        assert self.message_counter not in self.messages_to_send
        sequence_number = self.message_counter
        encoded_topic = topic.encode("utf-8")

        rbody = BytesIO()
        rbody.write(PublicationJournal.encode_int(sequence_number))
        rbody.write(PublicationJournal.encode_int(len(encoded_topic)))  # topic name length
        rbody.write(encoded_topic)  # topic name
        rbody.write(PublicationJournal.encode_int(len(message)))  # message size
        rbody.write(message)  # message data
        if key is not None:
            key = _ensure_bytes_like(key)
            rbody.write(PublicationJournal.encode_int(len(key)))  # key length
            rbody.write(key)  # key
        else:
            rbody.write(PublicationJournal.encode_int(0))  # no key
        if headers is not None:
            rbody.write(PublicationJournal.encode_int(len(headers)))  # number of headers
            for header in headers:
                # data must be encoded to be written
                hkey = _ensure_bytes_like(header[0])
                value = _ensure_bytes_like(header[1])
                rbody.write(PublicationJournal.encode_int(len(hkey)))  # header key length
                rbody.write(hkey)  # header key
                rbody.write(PublicationJournal.encode_int(len(value)))  # header value length
                rbody.write(value)  # header value
        else:
            rbody.write(PublicationJournal.encode_int(0))  # no headers

        self._write_record(PublicationJournal.msg_record_type_v2, rbody.getvalue())

        self.messages_to_send[sequence_number] = (message, topic, headers, key)
        self.message_counter += 1
        return sequence_number

    def has_messages_to_send(self):
        """Check whether there are messages queued for sending (which have either not been sent at
            all, or for which all sending attempts so far have failed, causing them to be requeued).
        """
        return len(self.messages_to_send) > 0

    def get_next_message_to_send(self):
        """Fetch the next message which should be sent

        Returns:
            The next message in the form of a tuple of
            (seqeunce number, message, target topic, message headers, message key),
            or None if there are no messages currently needing to be sent.
        """
        if len(self.messages_to_send) == 0:
            return None
        result = self.messages_to_send.pop_highest_priority()
        self.maybe_sent_messages[result[0]] = result[1]
        topic = result[1][1]
        if topic in self.inflight_topics:
            self.inflight_topics[topic] += 1
        else:
            self.inflight_topics[topic] = 1
        # rearrange into a single, new tuple, with the sequence number prepended
        return (result[0], *result[1])

    def has_messages_in_flight(self):
        """Check whether there are messages for which a sending attempt has been started,
            but has not yet conclusively succeeded or failed
        """
        return len(self.maybe_sent_messages) > 0

    def topics_with_messages_in_flight(self):
        """Get the set of topic names for which at least one message is currently in flight
        """
        return set(self.inflight_topics.keys())

    def mark_message_sent(self, sequence_number):
        """Mark a message as successfully sent, and removes it from further consideration.

        Truncates and restarts the backing journal file if the number of messages in-flight and
        waiting to be sent falls to zero, and restarts the sequence number assignment sequence.

        Raises:
            RuntimeError: If no message with the specifed sequence number is currently recorded as
                being in-flight.
        """
        if sequence_number not in self.maybe_sent_messages:
            raise RuntimeError(f"No record of message with sequence number {sequence_number} "
                               "being queued for sending")
        self._write_record(PublicationJournal.sent_record_type,
                           PublicationJournal.encode_int(sequence_number))
        topic = self.maybe_sent_messages[sequence_number][1]
        del self.maybe_sent_messages[sequence_number]
        if topic not in self.inflight_topics:
            raise Exception(f"Message to topic {topic} completed sending, but no messages were "
                            "known to be in flight to that topic")
        else:
            self.inflight_topics[topic] -= 1
            if self.inflight_topics[topic] == 0:
                del self.inflight_topics[topic]
        if len(self.messages_to_send) == 0 and len(self.maybe_sent_messages) == 0:
            # take this opportunity to garbage collect the journal file
            logger.debug("Journal is empty; garbage collecting")
            self.journal.close()
            os.unlink(self.journal_path)
            self.journal = open(self.journal_path, "wb")
            self.message_counter = 0

    def requeue_message(self, sequence_number):
        """Record a message send attempt as having failed by moving the message back from the
            in-flight pool to the queue of messages needing to be sent.

        Raises:
            RuntimeError: If no message with the specifed sequence number is currently recorded as
                being in-flight.
        """
        if sequence_number not in self.maybe_sent_messages:
            raise RuntimeError("No record of message with sequence number "
                               f"{sequence_number} being queued for sending")
        # nothing to record in journal, message was already written when first queued
        message = self.maybe_sent_messages[sequence_number]
        del self.maybe_sent_messages[sequence_number]
        self.inflight_topics[message[1]] -= 1
        if self.inflight_topics[message[1]] == 0:
            del self.inflight_topics[message[1]]
        self.messages_to_send[sequence_number] = message

    class _ReadPosition:
        """A type for tracking logical read positions in a journal file."""

        def __init__(self):
            self.record_index = 0
            self.record_start = 0

        def __str__(self):
            return f"record {self.record_index} which began at file offset {self.record_start}"

    @staticmethod
    def _read_raw_from_journal(journal, pos_info, size, name, allow_eof=False):
        """Read a specified size of raw data from a stream, producing an error if it cannot be
            fully read and end of file is not indicated to be tolerable.

        Args:
            journal: the stream from which to read
            pos_info: logical position information used for error messages
            size: the amount of data to read
            name: description of the data to be read, for error messages
            allow_eof: whether end of file is tolerable

        Returns:
            The block of data which was read

        Raises:
            RuntimeError: If the specified amount of data was not successfully read.
        """
        try:
            buffer = journal.read(size)
        except Exception:
            raise RuntimeError(f"Journal corrupted: Unable to read {name}")
        if len(buffer) == 0 and allow_eof:  # end of file
            return buffer, True
        elif len(buffer) < size:
            raise RuntimeError("Journal corrupted: Unexpected end of file (unable to read "
                               f"{name}) during {pos_info}")
        if allow_eof:
            return buffer, False
        return buffer

    @staticmethod
    def _decode_raw_data(data, decoder, crc, name):
        """Decode a block of raw data and update a running CRC.

        Args:
            data: the raw data to decode
            decoder: the specific decoding function to apply
            crc: the old value of the running CRC
            name: a description of the data for error messages

        Returns:
            A tuple of the result of applying the decoder and the updated CRC

        Raises:
            RuntimeError: If the decoder produces an error
        """
        crc = zlib.crc32(data, crc) & 0xFFFFFFFF
        try:
            decoded = decoder(data)
            return decoded, crc
        except BaseException as e:
            raise RuntimeError(f"Failed to decode {name}: {e}")

    @staticmethod
    def _read_recorded_header(journal, pos_info, bcrc, required_body_len, rec_len):
        """Extract a recorrded message header from a stream.

        Args:
            journal: the stream from which to read
            pos_info: logical position information used for error messages
            bcrc: the CRC of the message body as read so far
            required_body_len: the length the message record body must be to accomodate all
                               claimed data
            rec_len: the length of the message record body claimed in the record header

        Returns:
            A tuple consisting of the header key, header value, updated record body CRC, and
            updated required record body length.

        Raises:
            RuntimeError: If any read operation fails, data item cannot be decoded, or the supposed
                          header data does not fit inside the claimed record body length.
        """
        decode_int = PublicationJournal.decode_int
        int_size = PublicationJournal.int_size
        read = PublicationJournal._read_raw_from_journal
        decode = PublicationJournal._decode_raw_data

        raw_key_len = read(journal, pos_info, int_size, "message header key length")
        key_len, bcrc = decode(raw_key_len, decode_int, bcrc,
                               "message header key length")
        if key_len > (rec_len - required_body_len):
            raise RuntimeError(f"Journal corrupted: Claimed message header key length "
                               f"({key_len}) exceeds remaining space in record body for"
                               f" {pos_info}")
        required_body_len += key_len

        key = read(journal, pos_info, key_len, "message header key data")
        bcrc = zlib.crc32(key, bcrc) & 0xFFFFFFFF

        raw_val_len = read(journal, pos_info, int_size, "message header value length")
        val_len, bcrc = decode(raw_val_len, decode_int, bcrc,
                               "message header value length")
        if val_len > (rec_len - required_body_len):
            raise RuntimeError(f"Journal corrupted: Claimed message header value length"
                               f" ({val_len}) exceeds remaining space in record body "
                               f"for {pos_info}")
        required_body_len += val_len

        val = read(journal, pos_info, val_len, "message header value data")
        bcrc = zlib.crc32(val, bcrc) & 0xFFFFFFFF

        # confluent kafka requires keys in the form of unicode strings
        try:
            key = key.decode("utf-8")
        except UnicodeError:
            raise RuntimeError(f"Journal corrupted: Message header key is not valid "
                               f"UTF-8 in {pos_info}")
        return (key, val, bcrc, required_body_len)

    def _read_message_v1(self, journal, rpos, rec_len, bcrc):
        read = PublicationJournal._read_raw_from_journal
        decode = PublicationJournal._decode_raw_data
        read_header = PublicationJournal._read_recorded_header
        decode_int = PublicationJournal.decode_int

        # record must contain sequence number, message data len, number of message headers
        required_body_len = 3 * self.int_size
        if rec_len < required_body_len:
            raise RuntimeError(f"Journal corrupted: Claimed record body length ({rec_len}) "
                               f"too small to conain required data for {rpos}")

        raw_seq_num = read(journal, rpos, self.int_size, "message sequence number")
        seq_num, bcrc = decode(raw_seq_num, decode_int, bcrc, "sequence number")

        raw_msg_len = read(journal, rpos, self.int_size, "record message length")
        msg_len, bcrc = decode(raw_msg_len, decode_int, bcrc, "message length")
        if msg_len > (rec_len - required_body_len):
            raise RuntimeError(f"Journal corrupted: Claimed message data length ({msg_len})"
                               f" exceeds record body length for {rpos}")
        required_body_len += msg_len

        message_data = read(journal, rpos, msg_len, "record message data")
        bcrc = zlib.crc32(message_data, bcrc) & 0xFFFFFFFF

        raw_msg_hdr_cnt = read(journal, rpos, self.int_size, "record message header count")
        msg_hdr_cnt, bcrc = decode(raw_msg_hdr_cnt, decode_int, bcrc,
                                   "message header count")
        # each header is described by a minimum of two encoded integers (key len and value
        # len), so sanity check that the claimed number of headers would fit in the
        # remaining space in the record
        if 2 * self.int_size * msg_hdr_cnt > (rec_len - required_body_len):
            raise RuntimeError(f"Journal corrupted: Claimed number of message headers "
                               f"({msg_hdr_cnt}) exceeds remaining space in record body for {rpos}")
        required_body_len += 2 * self.int_size * msg_hdr_cnt

        message_headers = []
        for i in range(0, msg_hdr_cnt):
            key, val, bcrc, required_body_len = \
                read_header(journal, rpos, bcrc, required_body_len, rec_len)
            message_headers.append((key, val))

        if seq_num in self.messages_to_send:
            raise RuntimeError("Journal corrupted: Duplicate message sequence number in {rpos}")
        # Note that topic name and key are None because these were not stored in v1
        self.messages_to_send[seq_num] = (message_data, None, message_headers, None)
        if self.message_counter <= seq_num:
            self.message_counter = seq_num + 1

        return bcrc

    def _read_message_v2(self, journal, rpos, rec_len, bcrc):
        read = PublicationJournal._read_raw_from_journal
        decode = PublicationJournal._decode_raw_data
        read_header = PublicationJournal._read_recorded_header
        decode_int = PublicationJournal.decode_int

        # record must contain sequence number, topic length, message data length, key length,
        # and number of message headers
        required_body_len = 5 * self.int_size
        if rec_len < required_body_len:
            raise RuntimeError(f"Journal corrupted: Claimed record body length ({rec_len}) "
                               f"too small to conain required data for {rpos}")

        raw_seq_num = read(journal, rpos, self.int_size, "message sequence number")
        seq_num, bcrc = decode(raw_seq_num, decode_int, bcrc, "sequence number")

        raw_tpc_len = read(journal, rpos, self.int_size, "topic name length")
        tpc_len, bcrc = decode(raw_tpc_len, decode_int, bcrc, "topic name length")
        if tpc_len > (rec_len - required_body_len):
            raise RuntimeError(f"Journal corrupted: Claimed topic name length ({tpc_len})"
                               f" exceeds record body length for {rpos}")
        required_body_len += tpc_len

        raw_topic_name = read(journal, rpos, tpc_len, "topic name data")
        bcrc = zlib.crc32(raw_topic_name, bcrc) & 0xFFFFFFFF

        try:
            topic_name = raw_topic_name.decode("utf-8")
        except UnicodeError:
            raise RuntimeError(f"Journal corrupted: Topic name is not valid UTF-8 text in {rpos}")

        raw_msg_len = read(journal, rpos, self.int_size, "record message length")
        msg_len, bcrc = decode(raw_msg_len, decode_int, bcrc, "message length")
        if msg_len > (rec_len - required_body_len):
            raise RuntimeError(f"Journal corrupted: Claimed message data length ({msg_len})"
                               f" exceeds record body length for {rpos}")
        required_body_len += msg_len

        message_data = read(journal, rpos, msg_len, "record message data")
        bcrc = zlib.crc32(message_data, bcrc) & 0xFFFFFFFF

        raw_key_len = read(journal, rpos, self.int_size, "key length")
        key_len, bcrc = decode(raw_key_len, decode_int, bcrc, "key length")
        if key_len > (rec_len - required_body_len):
            raise RuntimeError(f"Journal corrupted: Claimed key length ({key_len})"
                               f" exceeds record body length for {rpos}")
        required_body_len += key_len

        if key_len > 0:
            key = read(journal, rpos, key_len, "key data")
            bcrc = zlib.crc32(key, bcrc) & 0xFFFFFFFF
        else:
            key = None

        raw_msg_hdr_cnt = read(journal, rpos, self.int_size, "record message header count")
        msg_hdr_cnt, bcrc = decode(raw_msg_hdr_cnt, decode_int, bcrc,
                                   "message header count")
        # each header is described by a minimum of two encoded integers (key len and value
        # len), so sanity check that the claimed number of headers would fit in the
        # remaining space in the record
        if 2 * self.int_size * msg_hdr_cnt > (rec_len - required_body_len):
            raise RuntimeError(f"Journal corrupted: Claimed number of message headers "
                               f"({msg_hdr_cnt}) exceeds remaining space in record body for {rpos}")
        required_body_len += 2 * self.int_size * msg_hdr_cnt

        message_headers = []
        for i in range(0, msg_hdr_cnt):
            hkey, val, bcrc, required_body_len = \
                read_header(journal, rpos, bcrc, required_body_len, rec_len)
            message_headers.append((hkey, val))

        if seq_num in self.messages_to_send:
            raise RuntimeError("Journal corrupted: Duplicate message sequence number in {rpos}")
        self.messages_to_send[seq_num] = (message_data, topic_name, message_headers, key)
        if self.message_counter <= seq_num:
            self.message_counter = seq_num + 1

        return bcrc

    def _read_previous_journal(self):
        """Reload journal data from a previous session from disk.

        Raises:
            PermissionError: If the journal file does not have suitable permissions.
            RuntimeError: If a problem occurs reading or interpreting the stored journal data.
        """
        if not os.path.exists(self.journal_path):
            return  # nothing to do!
        journal = open(self.journal_path, "rb")
        rpos = PublicationJournal._ReadPosition()
        read = PublicationJournal._read_raw_from_journal
        decode = PublicationJournal._decode_raw_data

        while True:
            decode_int = PublicationJournal.decode_int
            decode_crc = PublicationJournal.decode_crc

            rpos.record_start = journal.tell()

            # Read the record header:
            hcrc = 0  # header CRC
            record_type_offset = journal.tell()
            raw_rec_type, eof = read(journal, rpos, self.int_size, "record type", allow_eof=True)
            if eof:
                break
            rec_type, hcrc = decode(raw_rec_type, decode_int, hcrc, "record type")

            raw_rec_len = read(journal, rpos, self.int_size, "record body length")
            rec_len, hcrc = decode(raw_rec_len, decode_int, hcrc, "record body length")

            raw_orig_body_crc = read(journal, rpos, self.crc_size, "record body CRC")
            orig_body_crc, hcrc = decode(raw_orig_body_crc, decode_crc, hcrc, "record body CRC")

            raw_orig_hdr_crc = read(journal, rpos, self.crc_size, "record header CRC")
            # don't update recalculated CRC with raw_orig_hdr_crc, since it did not include itself
            orig_hdr_crc, _ = decode(raw_orig_hdr_crc, decode_crc, 0, "record header CRC")

            if hcrc != orig_hdr_crc:
                raise RuntimeError(f"Journal corrupted: Header CRC mismatch (original: "
                                   f"{orig_hdr_crc:x}, recalculated: {hcrc:x}) for {rpos}")

            # Read the record body:
            bcrc = 0  # body CRC

            if rec_type == PublicationJournal.msg_record_type_v1:
                bcrc = self._read_message_v1(journal, rpos, rec_len, bcrc)

            elif rec_type == PublicationJournal.msg_record_type_v2:
                bcrc = self._read_message_v2(journal, rpos, rec_len, bcrc)

            elif rec_type == PublicationJournal.sent_record_type:
                raw_seq_num = read(journal, rpos, self.int_size, "message sequence number")
                seq_num, bcrc = decode(raw_seq_num, decode_int, bcrc, "sequence number")

                if seq_num in self.messages_to_send:
                    del self.messages_to_send[seq_num]
                else:
                    raise RuntimeError("Journal corrupted: Record of sent message "
                                       f"({seq_num}) which did not previously appear, in {rpos}")
            else:
                raise RuntimeError(f"Journal corrupted: Invalid record type ({rec_type}) "
                                   f"at file offset {record_type_offset}")

            if bcrc != orig_body_crc:
                raise RuntimeError(f"Journal corrupted: CRC mismatch (original: {orig_body_crc:x}, "
                                   f"recalculated: {bcrc:x}) for {rpos}")

            rpos.record_index += 1

    class NullLock:
        """A trivial context manager-compatible class which can be used in place of a lock when no
        locking is needed."""

        def __init__(self):
            pass

        def __enter__(self):
            pass

        def __exit__(self, exc_type, exc_value, exc_traceback):
            pass

    def _delivery_callback(self,
                           kafka_error: confluent_kafka.KafkaError,
                           msg: confluent_kafka.Message,
                           seq_num: int,
                           lock):
        """Handle a callback from librdkafka indicating the outcome of sending a message by either
            marking it successfully sent or requeuing it to send again.
        """
        if kafka_error is None and msg.error() is None:
            logger.debug(f"Message with sequence number {seq_num} was confirmed sent in "
                         f"{msg.latency()} seconds")
            with lock:
                self.mark_message_sent(seq_num)
        else:
            err = kafka_error
            if err is None:
                err = msg.error()
            if err.retriable():
                logger.error("Error delivering message with sequence number: %i: "
                             "%s; requeuing to send again", seq_num, err)
                with lock:
                    self.requeue_message(seq_num)
            else:
                logger.error("Error delivering message with sequence number: %i: "
                             "%s; error is not retriable so message cannot be sent", seq_num, err)
                with lock:
                    self.mark_message_sent(seq_num)

    def get_delivery_callback(self, seq_num, lock=NullLock()):
        """Construct a callback handler specific to a particular message which will either mark it
        successfully sent or requeue it to send again.

        The callback which is produced will take two arguments: A confluent_kafka.KafkaError
        describing any error in sending the message, and confluent_kafka.Message containing the
        message itself.

        Args:
            seq_num: The sequence number of the message in question, previously returned by
                     :meth:`get_next_message_to_send`.
            lock: An optional reference to a lock object which the callback should hold when
                  invoked, e.g. to protect concurrent access to the journal.

        """
        if seq_num not in self.maybe_sent_messages:
            raise RuntimeError("Cannot produce a delivery callback for message with sequence "
                               f"number {seq_num} which is not in flight")
        return lambda err, msg: self._delivery_callback(err, msg, seq_num, lock)

    # adc.errors.log_client_errors is dangerous as it throws exceptions
    # which will trigger https://github.com/confluentinc/confluent-kafka-python/issues/729
    @staticmethod
    def error_callback(kafka_error: confluent_kafka.KafkaError):
        """A safe callback handler for reporting Kafka errors."""
        logger.error(f"Kafka error: {kafka_error}")


class RobustProducer(threading.Thread):
    def __init__(self, url, auth=True, journal_path="publisher.journal", poll_wait=1.e-4, **kwargs):
        """Construct a publisher which will retry sending messages if it does not receive
        confirmation that they have arrived, including if it is itself taken offline (i.e. crashes)
        for some reason.

        This is intended to provide *at least once* delivery of messages: If a message is confirmed
        received by the broker, it will not be sent again, but if any disruption of the network or
        the publisher itself prevents it from receiving that confirmation, even if the message was
        actually received by the broker, the publisher will assume the worst and send the message
        again. Users of this class (and more generally consumers of data published with it) should
        be prepared to discard duplicate messages.

        Args:
            url: The URL for the Kafka topci to which messages will be published.
            auth: A `bool` or :class:`Auth <hop.auth.Auth>` instance. Defaults to
                  loading from :meth:`auth.load_auth <hop.auth.load_auth>` if set to
                  True. To disable authentication, set to False.
            journal_path: The path on the filesystem where the messages being sent should be
                          recorded until they are known to have been successfully received. This
                          path should be located somewhere that will survive system restarts, and if
                          messages contain sensitive data it should be noted that they will be
                          written unencrypted to this path. The journal size is generally limited to
                          the sum of sizes of messages queued for sending or in flight at the same
                          time, plus some small (few tens of bytes per message) bookkeeping
                          overhead. Note that this size can become large is a lengthy network
                          disruption prev ents messages from being sent; enough disk spacec should
                          be available to cover this possibility for the expected message rate and
                          duration of disruptions which may need to be handled.
            poll_wait: The time the publisher should spend checking for receipt of each message
                       directly after sending it. Tuning this parameter controls a tradeoff between
                       low latency discovery of successful message delivery and throughput. If the
                       time between sending messages is large compared to the latency for a message
                       to be sent and for a confirmation of receipt to return, it is useful to
                       increase this value so that the publisher will wait to discover that each
                       message has been sent (in the success case) instead of sleeping and waiting
                       for another message to send. If this value is 'too low' (much smaller than
                       both the time for a message to be sent and acknowledged and the time for the
                       next message to be ready for sending), the publisher will waste CPU time
                       entering and exiting the internal function used to receive event
                       notifications. If this value is too large (larger than or similar in size to
                       the time between messages needing to be sent) throughput will be lost as time
                       will be spent waiting to see if the previous message has been acknowledged
                       which could be better spent getting the next message sent out. When in doubt,
                       it is probably best to err on the side of choosing a small value.
            kwargs: Any additional arguments to be passed to :meth:`hop.io.open <hop.io.open>`.

        Raises:
            OSError: If a journal file exists but cannot be read.
            Runtime Error: If the contents of the journal file are corrupted.

        """
        super(RobustProducer, self).__init__()

        self._poll_wait = poll_wait

        # read any messages left over from a previous run
        self._journal = PublicationJournal(journal_path)
        if self._journal.has_messages_to_send():
            logger.info(f"Journal has {len(self._journal.messages_to_send)} "
                        "left-over messages to send")

        self._lock = threading.Lock()
        self._cond = threading.Condition(self._lock)

        dummy = io.Stream(auth=auth)
        if dummy.auth is not None:
            username, broker_addresses, _ = kafka.parse_kafka_url(url)
            self._credential = select_matching_auth(dummy.auth, broker_addresses[0], username)
        else:
            self._credential = None
        self._stream = dummy.open(url, "w", error_callback=PublicationJournal.error_callback,
                                  **kwargs)

    def run(self):
        """This method is not part of the public interface of this class, and should not be called
        directly by users.

        """
        while True:
            with self._cond:
                if self._should_stop and not self._journal.has_messages_to_send():
                    break
                if not self._immediate_start:
                    self._cond.wait()  # wait for something to do
                else:
                    self._immediate_start = False
            self._do_send()
        self._stream.flush()

    def _do_send(self):
        journal = self._journal
        # use this slightly awkward loop structure to reacquire the lock on each iteration so that
        # other threads have a chance to take it
        while True:
            with self._lock:
                busy = journal.has_messages_to_send() or journal.has_messages_in_flight()
                if not busy:
                    break  # no work to do for now
                do_send = False
                if journal.has_messages_to_send():
                    seq_num, message, topic, headers, key = journal.get_next_message_to_send()
                    do_send = True
            if do_send:
                try:
                    dc = journal.get_delivery_callback(seq_num, self._lock)
                    logger.debug(f"Sending message with sequence number {seq_num}")
                    self._stream.write_raw(message, headers=headers, delivery_callback=dc,
                                           topic=topic, key=key)
                except KafkaException as e:
                    logger.error(f"Error sending message with sequence number: {seq_num}: {e}"
                                 "; requeuing to send again")
                    with self._lock:
                        journal.requeue_message(seq_num)
            try:
                # if we just sent a message, do a full-length wait on that topic
                if do_send:
                    rec = self._stream._record_for_topic(topic)
                    rec.producer._producer.poll(self._poll_wait)
                else:
                    # if we did not send, then there was at least one message already in flight, and
                    # no messages queued to send, so wait to see if any of the in-flight messages
                    # get delivered
                    time.sleep(self._poll_wait)
                # Check all topics with in-flight messages for deliveries.
                # Wait for 0 time on each individually because we already did a wait above
                with self._lock:
                    inflight_topics = journal.topics_with_messages_in_flight()
                for topic in inflight_topics:
                    # This is sub-optimal because more than one topic may map to the same producer,
                    # which can cause a producer to be visited multiple times by this loop
                    rec = self._stream._record_for_topic(topic)
                    rec.producer._producer.poll(0)
            except KafkaException:
                pass

    def write(self, message, headers=None, test=False, topic=None, key=None):
        """Queue a message to be sent. Message sending occurs asynchronously on a background thread,
        so this method returns immediately unless an error occurs queuing the message.
        :meth:`RobustProducer.start <RobustProducer.start>` must be called prior to calling this
        method.

        Args:
            message: A message to send.
            headers: Headers to be sent with the message, as a list of 2-tuples of strings.
            test: The message should be marked as a test message by adding a header with
                  key '_test'.
            topic: The topic to which the message should be sent. This need not be specified if the
                   producer was constructed with a URL containing exactly one topic name.
            key: If specified, the Kafka message key

        Raises:
            RuntimeError: If appending the new message to the on-disk journal fails.
            TypeError: If the message is not a suitable type.
            Exception: If a single topic was not specified in the original URL and the topic
                       argument was not set.

        """
        if topic is None and self._stream.default_topic is not None:
            topic = self._stream.default_topic
        if topic is None:
            raise Exception("No topic specified for write: "
                            "Either configure a topic when opening the Stream, "
                            "or specify the topic argument to write()")
        message, headers = io.Producer.pack(message, headers, auth=self._credential, test=test)
        with self._cond:  # must hold the lock to manipulate journal
            seq_num = self._journal.queue_message(message, topic, headers, key=key)
            self._cond.notify()  # wake up the sender loop if sleeping
        logger.debug(f"Queued message with sequence number {seq_num} to be sent")

    def start(self):
        """Start the background communication thread used by the publisher to send messages. This
        should be called prior to any calls to :meth:`RobustProducer.write <RobustProducer.write>`.
        This method should not be called more than once.

        """
        self._should_stop = False
        self._immediate_start = self._journal.has_messages_to_send()
        super(RobustProducer, self).start()

    def stop(self):
        """Stop the background communication thread used by the publisher to send messages. This
        method will block until the thread completes, which includes sending all queued messages.
        :meth:`RobustProducer.write <RobustProducer.write>` should not be called after this method
        has been called.
        This method should not be called more than once.

        """
        logger.debug("Stopping publisher thread")
        with self._cond:
            self._should_stop = True
            self._cond.notify()  # wake up the sender loop
        self.join()
        self._stream.close()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, ex_type, ex_val, traceback):
        self.stop()
        return False  # propagate any exceptions



from contextlib import redirect_stdout
import io
import os
from pathlib import Path
import sys

import pytest

from hop import models


# test the subscribe printer for each message format
@pytest.mark.parametrize("message_format", ["voevent", "circular", "blob", "json", "avro"])
def test_print_message(message_format, message_parameters_dict):
    # load parameters from conftest
    message_parameters = message_parameters_dict[message_format]
    model_name = message_parameters["model_name"]
    test_file = message_parameters["test_file"]

    shared_datadir = Path("tests/data")

    test_content_path = shared_datadir / "test_data" / test_file

    # load model
    model_class = message_parameters["expected_model"]
    if issubclass(model_class, models.MessageModel):
        model = model_class.load_file(test_content_path)
    else:
        model = test_content_path.read_text()

    f = io.StringIO()
    with redirect_stdout(f):
        print(model, file=sys.stdout, flush=True)

    # extract message string from stdout
    test_message_stdout_str = f.getvalue()
    print(test_message_stdout_str)

    # read in expected stdout text
    expected_basename = os.path.splitext(test_file)[0]
    expected_file = expected_basename + ".stdout"
    expected_message_stdout = (shared_datadir / "expected_data" / expected_file).read_text()

    # verify printed message structure is correct
    assert test_message_stdout_str == expected_message_stdout

    f.close()



from importlib.metadata import version
import confluent_kafka


def print_packages_versions():
    """Print versions for the passed packages.

    """
    packages = get_packages()
    for pkg in packages:
        if pkg == "librdkafka":
            print("%s==%s" % (pkg, confluent_kafka.libversion()[0]))
        else:
            print("%s==%s" % (pkg, version(pkg)))


def get_packages():
    """Returns the package dependencies used within hop-client.

    """
    return ["hop-client", "adc_streaming", "confluent_kafka", "librdkafka"]


def _main(args):
    """List all the dependencies' versions.

    """

    print_packages_versions()
